/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * %DelayLine class template.
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#ifndef DELAYLINE_H
#define DELAYLINE_H

#include <cassert>  // for assert()
#include <memory>   // for std::allocator
// TODO: remove output functions
#include <iostream> // for output
#include <cmath>    // for cos()
#include <vector>   // for std::vector
#include <cstring>  // for memcpy()

#include "classtools.h"
#include "ssr_global.h"

/** Delay line.
 * If @a initial_delay > 0 then negative indexing can be used (of course only
 * up to @a -initial_delay).
 *
 * @deprecated The @e read @e position is set by the write() method. Therefore all read()
 * operations are relative to the most recently written block. I guess this is
 * straightforward functionality, isn't it?
 * 
 * @warning This was changed recently, now the @a _read_reference is set by the
 * member function prepare_write(), which has to be called before the first
 * write operation.
 *
 * @todo allowed values for sizes and delays are checked on each read/write, it
 * may be cheaper (but less secure?) to check those things in the calling
 * function.
 * @warning A restriction to the type @b T is that it must allow assignment of 0
 * to it. Most of the basic types should be ok with that anyway. Pointers to
 * anything also work. std::string@e s don't work (as far as I know).
 * @warning There is another restriction on the type @b T as soon as
 * interpolation is used. In this case only @b float related types will work.
 */
template <typename T>
class DelayLine : classtools::NonCopyable
{
  public:
    /// Define @a size_type dependent on the used type
    typedef typename std::allocator<T>::size_type size_type;
    typedef typename std::allocator<T>::difference_type difference_type;

    /** ctor.
     * @param size the size (in samples) of the delay line
     * @param initial_delay global delay added to any read operation. This
     * makes negative indexing possible. (It is needed for focussed point
     * sources). If you don't need that, just leave it at the default 0.
     * @warning If @a initial_delay >= @a size you will get a problem! This is
     * not checked.
     **/
    explicit DelayLine(const size_type size, const size_type initial_delay=0) :
      _size(size),
      _write_position(0),
      _read_reference(0),
      _initial_delay(initial_delay)
    {
      // TODO: new can throw exception!
      _buffer = new T[size];
      // initialize buffer[] with zeros.
      for (size_type i = 0; i < _size; ++i) _buffer[i] = 0;
    }

    /** dtor.
     * Frees the reserved memory.
     **/
    ~DelayLine()
    {
      delete [] _buffer;
    }

    // for the implementation of the following methods, see end of file
    bool write(const T data[], size_type data_size);
    void prepare_write();
    bool read(T data[], size_type data_size, int delay = 0) const;
    bool read_mult(T data [], size_type data_size, int delay = 0,
        float factor = 1.0, bool accumulate = false) const;
    bool read_fade(T data [], size_type data_size, int delay1, float factor1,
        int delay2, float factor2, bool accumulate) const;
    bool read_interp(T data [], size_type data_size, int delay1, float factor1,
         int delay2, float factor2, bool accumulate) const;

  private:
    /// Pointer to stored data
    T* _buffer;
    /// length of delay line (it is set in the constructor and stays constant)
    const size_type _size;
    /// %Position where data will be written to
    size_type _write_position;
    /// points to the first sample of the current block
    size_type _read_reference;

    /** %Global delay applied to the delayline.
     * this makes negative indexing possible (for focussed sources)
     * (it is set in the constructor and stays constant)
     **/
    const size_type _initial_delay;

    mutable std::vector<T> _fade_in, _fade_out; // for fading
};

/** Write a block of data into the delay line.
 * The read-pointer is set to the beginning of the block - @a initial_delay
 * The write-pointer is set to just after the end of the written data.
 * @param data an array of data to be written
 * @param data_size size of this array
 * @return @b true if successful.
 **/
template <typename T>
bool DelayLine<T>::write(const T data[], const size_type data_size)
{
  if (data_size + _initial_delay > _size) return false;
  size_type tail_space = _size - _write_position;
  if (data_size > tail_space)
  {
    memcpy(&_buffer[_write_position], &data[0], tail_space * sizeof(T));
    memcpy(&_buffer[0], &data[tail_space], (data_size-tail_space) * sizeof(T));
  }
  else
  {
    assert(data != NULL);
    memcpy(&_buffer[_write_position], &data[0], data_size * sizeof(T));
  }
  _write_position += data_size;
  if (_write_position >= _size) _write_position %= _size;
  return true;
}

/** Quick-Hack to fix a nasty bug.
 * This function should be replaced soon!
 * A reimplementation of the DelayLine class is planned anyway where this
 * problem should be solved <em>avec plus d'elegance</em> ...
 * This function must be called before the first write operation of the current
 * audio buffer!
 **/
template <typename T>
void DelayLine<T>::prepare_write()
{
  if (_initial_delay > _write_position)
  {
    _read_reference = _write_position + _size - _initial_delay;
  }
  else
  {
    _read_reference = _write_position - _initial_delay;
  }
}

/** Read a block of data.
 * @param data good old C array which will hold the data
 * @param data_size the size of this array
 * @param delay the actual delay in samples (default 0)
 * @return @b true if successful. If @b false, a negative value for @a
 * delay exceeded the @a initial_delay, @b or @a delay exceeded the
 * @a size of the delay line, @b or @a data_size was too big.
 * @warning If @b false is returned, @a data doesn't contain any usable values!
 * @warning If the last writing buffer and the reading buffer have different
 * sizes, there can be problems, which are not checked for! In our case there is
 * no need because the write and read operations are always in the same JACK
 * audio cycle and therefore have the same blocksize.
 * @bug It is very likely that there are bugs, so I'm already reserving this
 * space!
 **/
template <typename T>
bool DelayLine<T>::read(T data[], const size_type data_size,
    const int delay) const
{
  if (delay < -static_cast<int>(_initial_delay))
  {
    std::cerr << "Negative delay is larger than initial_delay!" << std::endl;
    return false;
  }
  if (delay > static_cast<int>(_size - data_size - _initial_delay))
  {
    return false;
  }

  difference_type read_position = _read_reference - delay;
  while (read_position < 0) read_position += _size;

  // size_type read_position;
  // if (delay > static_cast<int>(_read_reference))
  // {
  //   read_position = _read_reference + _size - delay;
  // }
  // else
  // {
  //   read_position = _read_reference - delay;
  // }
  // if (read_position >= _size) read_position %= _size;

  size_type tail_space = _size - read_position;
  if (data_size > tail_space)
  {
    memcpy(&data[0], &_buffer[read_position], tail_space * sizeof(T));
    memcpy(&data[tail_space], &_buffer[0], (data_size-tail_space) * sizeof(T));
  }
  else
  {
    memcpy(&data[0], &_buffer[read_position], data_size * sizeof(T));
  }
  return true;
}

/** Read a block of data and multiply every element by @a factor.
 * You can choose if the result is simply stored or if it's added to @a data.
 * @param data good old C array which will hold the data
 * @param data_size the size of this array
 * @param delay the actual delay in samples (default 0)
 * @param factor each element is multiplied by it (default 1.0)
 * @param accumulate if @b true, new data is added to existing @a data; if @b
 * false, new data replaces @a data.
 * @return @b true if successful. If @b false, a negative value for @a
 * delay exceeded the @a initial_delay, @b or @a delay exceeded the
 * @a size of the delay line, @b or @a data_size was too big.
 * @todo If @b false is returned, @a data hopefully still contains the old
 * values!
 * @warning If the last writing buffer and the reading buffer have different
 * sizes, there can be problems, which are not checked for! In our case there is
 * no need because the write and read operations are always in the same JACK
 * audio cycle and therefore have the same blocksize.
 * @bug It is very likely that there are bugs, so I'm already reserving this
 * space!
 **/
template <typename T>
bool DelayLine<T>::read_mult(T data [], const size_type data_size,
    const int delay, const float factor,
    const bool accumulate) const
{
  // copied from above! TODO: there should be another way to reuse the code!
  if (delay < -static_cast<int>(_initial_delay))
  {
    std::cerr << "Negative delay is larger than initial_delay!" << std::endl;
    return false;
  }
  if (delay > static_cast<int>(_size - data_size - _initial_delay))
  {
    return false;
  }

  difference_type read_position = _read_reference - delay;
  while (read_position < 0) read_position += _size;

  // size_type read_position;
  // if (delay > static_cast<int>(_read_reference))
  // {
  //   read_position = _read_reference + _size - delay;
  // }
  // else
  // {
  //   read_position = _read_reference - delay;
  // }
  // if (read_position >= _size) read_position %= _size;

  //////////////////////////////////////////////////////////////////

  size_type in = read_position;
  size_type out = 0;
  if (accumulate)
  {
    while (out < data_size)
    {
      data[out++] += _buffer[in++] * factor;
      if (in >= _size) in %= _size;
    }
  }
  else // don't accumulate
  {
    // TODO: nearly the same as above. Code reuse?
    while (out < data_size)
    {
      data[out++] = _buffer[in++] * factor;
      if (in >= _size) in %= _size;
    }
  }
  return true;
}

template <typename T>
bool DelayLine<T>::read_fade(T data [], const size_type data_size,
    const int delay1, const float factor1,
    const int delay2, const float factor2,
    const bool accumulate) const
{
  // copied from above! TODO: there should be another way to reuse the code!
  if (delay1 < -static_cast<int>(_initial_delay))
  {
    std::cerr << "Negative delay is larger than initial_delay!" << std::endl;
    return false;
  }
  if (delay1 > static_cast<int>(_size - data_size - _initial_delay))
  {
    return false;
  }

  difference_type read_position1 = _read_reference - delay1;
  while (read_position1 < 0) read_position1 += _size;
  if (read_position1 >= _size) read_position1 %= _size;

  // TODO: this is just here for future debugging. Remove eventually, or replace class  ;)
  //VERBOSE2("#### read_position1 = " << read_position1 << ", _read_reference = " << _read_reference << ", delay1 = " << delay1);
  

  // size_type read_position1;
  // if (delay1 > static_cast<int>(_read_reference))
  // {
  //   read_position1 = _read_reference + _size - delay1; // BUG?
  // }
  // else
  // {
  //   read_position1 = _read_reference - delay1;
  // }
  // if (read_position1 >= _size) read_position1 %= _size;

  //////////////////////////////////////////////////////////////////
  // copied from above! TODO: there should be another way to reuse the code!
  if (delay2 < -static_cast<int>(_initial_delay))
  {
    std::cerr << "Negative delay is larger than initial_delay!" << std::endl;
    return false;
  }
  if (delay2 > static_cast<int>(_size - data_size - _initial_delay))
  {
    return false;
  }

  difference_type read_position2 = _read_reference - delay2;
  while (read_position2 < 0) read_position2 += _size;
  if (read_position2 >= _size) read_position2 %= _size;

  // size_type read_position2;
  // if (delay2 > static_cast<int>(_read_reference))
  // {
  //   read_position2 = _read_reference + _size - delay2;
  // }
  // else
  // {
  //   read_position2 = _read_reference - delay2;
  // }
  // if (read_position2 >= _size) read_position2 %= _size;

  ////////////////////////////////////////////////////////////////////

#undef PI
  const T PI = 3.14159265358979323846;

  if (_fade_in.size() != data_size)
  {
    // std::cout << "Resizing fade in to " << data_size << " samples.\n";
    _fade_in.resize(data_size);
    for (typename std::vector<T>::size_type i = 0; i < data_size; ++i)
    {
      _fade_in[i] = cos(i*PI/data_size)*-0.5+0.5;
    }
  }
  if (_fade_out.size() != data_size)
  {
    // std::cout << "Resizing fade out to " << data_size << " samples.\n";
    _fade_out.resize(data_size);
    for (typename std::vector<T>::size_type i = 0; i < data_size; ++i)
    {
      _fade_out[i] = cos(i*PI/data_size)*0.5+0.5;
    }
  }

  size_type in1 = read_position1;
  size_type in2 = read_position2;
  
  size_type out = 0;
  if (accumulate)
  {
    while (out < data_size)
    {
      // TODO: use look-up-table for fade in and fade out!
      // This ist just a temporary implementation

      // raised cosine window:
      data[out] += _buffer[in1++] * factor1 * _fade_out[out]
        + _buffer[in2++] * factor2 * _fade_in[out];
      // linear fade:
      /*
      data[out] += _buffer[in1++] * factor1 * out/(T)data_size
        + _buffer[in2++] * factor2 * (1.0 - out/(T)data_size);
        */
      ++out;
      if (in1 >= _size) in1 %= _size;
      if (in2 >= _size) in2 %= _size;
    }
  }
  else // don't accumulate
  {
    // TODO: nearly the same as above. Code reuse?
    while (out < data_size)
    {
      // TODO: use look-up-table for fade in and fade out!
      // This ist just a temporary implementation

      // raised cosine window:
      data[out] = _buffer[in1++] * factor1 * _fade_out[out]
        + _buffer[in2++] * factor2 * _fade_in[out];
      // linear fade:
      /*
      data[out] = _buffer[in1++] * factor1 * out/(T)data_size
        + _buffer[in2++] * factor2 * (1.0 - out/(T)data_size);
        */
      ++out;
      if (in1 >= _size) in1 %= _size;
      if (in2 >= _size) in2 %= _size;
    }
  }
  return true;
}

template <typename T>
bool DelayLine<T>::read_interp(T data [], const size_type data_size,
    const int prev_delay, const float prev_weight,
    const int delay, const float weight,
             const bool accumulate) const
{
  // save processing resources
  if (prev_delay == delay)
  {
    // return read_mult(data, data_size, delay, weight, accumulate);
    return read_fade(data, data_size, prev_delay, prev_weight, delay,
        weight, accumulate);
  }

  // copied from above! TODO: there should be another way to reuse the code!
  if (prev_delay < -static_cast<int>(_initial_delay))
  {
    std::cerr << "Negative delay is larger than initial_delay!" << std::endl;
    return false;
  }
  if (prev_delay > static_cast<int>(_size - data_size - _initial_delay))
  {
    return false;
  }

  size_type read_position;

  if (prev_delay > static_cast<int>(_read_reference))
  {
    read_position = _read_reference + _size - prev_delay;
  }
  else
  {
    read_position = _read_reference - prev_delay;
  }
  if (read_position >= _size) read_position %= _size;
  //////////////////////////////////////////////////////////////////
  // copied from above! TODO: there should be another way to reuse the code!
  if (delay < -static_cast<int>(_initial_delay))
  {
    std::cerr << "Negative delay is larger than initial_delay!" << std::endl;
    return false;
  }
  if (delay > static_cast<int>(_size - data_size - _initial_delay))
  {
    return false;
  }
  ////////////////////////////////////////////////////////////////////

  size_type in = read_position;
  size_type in_plus_one = read_position+1;

  // determine how many samples to read from the delay line
  //  const size_type size_to_read = delay - prev_delay + data_size;

  T current_audio_sample;
  float fractional_time = 0.0f; // in samples

  size_type x_0 = static_cast<size_type>(fractional_time);

#undef PI
  const T PI = 3.14159265358979323846;

  for (size_type write_counter = 0; write_counter < data_size; )
  {
    // round downwards
    x_0 = static_cast<size_type>(fractional_time);

    // determine reading positions
    in = (read_position + x_0) % _size;
    in_plus_one = (read_position + x_0 + 1) % _size;


    // linear interpolation: y = y_0 + (x - x_0) * (y_1 - y_0)/(x_1 - x_0)
    current_audio_sample = _buffer[in] + (fractional_time - x_0) * (_buffer[in_plus_one] - _buffer[in]);
    // linear interpolation of weights
    current_audio_sample *= prev_weight + write_counter * (weight - prev_weight)/data_size;

    // write to output buffer
    if (accumulate) data[write_counter] += current_audio_sample;
    // don't accumulate
    else data[write_counter] = current_audio_sample;

    // find factional position for next reading operation
    fractional_time += 1.0f + (delay - static_cast<float>(prev_delay))/static_cast<float>(data_size) *
      (cos(static_cast<float>(write_counter)/static_cast<float>(data_size) * 2.0f * PI -  PI) + 1.0f);

    write_counter++;
  } // for 
  return true;
}

#endif

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
