/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * %NewVBAPRenderer class (implementation).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#include <cassert>       // for assert()
#include <cmath>         // for cos()

#include "newvbaprenderer.h"
#include "controller.h"  // for set_active_sources(), ...
#include "ssr_global.h"
#include "maptools.h"    // for purge()
#include "mathtools.h"    // for purge()
#include "stringtools.h" // for A2S()

// for debugging only:
#include <libxml/debugXML.h>

using stringtools::A2S;
using mathtools::deg2rad;
using namespace ssr; // for id_t, processing, ready, ...

/** ctor.
 * @param scene see create()
 * @param controller see create()
 * @param delayline_size see create()
 **/
NewVBAPRenderer::NewVBAPRenderer(const Scene& scene, Controller& controller
    , const parameter_map &params)
  : NewRenderer(scene, controller, params) // base class ctor.
{
  if (!_make_outputs())
  {
    //throw(vbap_error("Output objects could not be created!"));
    throw(std::runtime_error("Output objects could not be created!"));
  }

  // There shouldn't be any sources in the scene at this point.
  // If there are, they are ignored!

  // fill silence buffer with zeroes
  _silence.resize(this->get_buffer_size(), 0);

  // init memory for crossfade slopes
  _fade_in.resize (this->get_buffer_size(), static_cast<sample_t>(0.0f));
  _fade_out.resize(this->get_buffer_size(), static_cast<sample_t>(0.0f));

  // raised cosine fade
  for (unsigned int n = 0u; n < this->get_buffer_size(); n++)
  {
    _fade_in[n]  = static_cast<sample_t>(0.5f
        + 0.5f * cos(static_cast<float>(this->get_buffer_size()-n)
          / static_cast<float>(this->get_buffer_size()) * mathtools::PI));
    _fade_out[n] = static_cast<sample_t>(0.5f
        + 0.5f * cos(static_cast<float>(n)
          / static_cast<float>(this->get_buffer_size()) * mathtools::PI));
  }
}

NewVBAPRenderer::~NewVBAPRenderer()
{}

// TODO: These are boilerplate, i could create them
//       with templates in the renderer.
//
//       not 100% sure though. the NewRenderer Class must not
//       be a template. so i guess, i need a template Class deriving from it.

NewRenderer::Input *
NewVBAPRenderer::create_input(const parameter_map &params)
{
  VBAPInput* dumbinput = new VBAPInput(*this, params);
  return (Input *) dumbinput;
}

NewRenderer::Output *
NewVBAPRenderer::create_output(const parameter_map &params)
{
  VBAPOutput* dumboutput = new VBAPOutput(*this, params);
  return (Output *) dumboutput;
}

NewRenderer::RenderSource *
NewVBAPRenderer::create_source(NewRenderer::Input *input
    , const parameter_map &params)
{
  VBAPSource* dumbsource = new VBAPSource(*this, input, params);
  return (RenderSource *) dumbsource;
}

void
NewVBAPRenderer::render(RenderSource *s, Output *o)
{
  //VBAPInput  *input  = dynamic_cast<VBAPInput *>(s->get_input());
  VBAPSource *source = dynamic_cast<VBAPSource *>(s);
  VBAPOutput *output = dynamic_cast<VBAPOutput *>(o);

  if (source->left_speaker == output)
  {
    _render_source_on_output(source, output, source->weight_left, _nframes);
    source->set_output_activity(output->id, true);
  }
  else if (source->right_speaker == output)
  {
    _render_source_on_output(source, output, source->weight_right, _nframes);
    source->set_output_activity(output->id, true);
  }
  else if (output->_speaker.model == Loudspeaker::subwoofer)
  {
    _render_source_on_output(source, output, 1.0, _nframes);
    source->set_output_activity(output->id, true);
  }
  else
  {
    source->set_output_activity(output->id, false);
  }
}

/** _.
 * This will only be called once. Loudspeakers can @b not be added later.
 * %Loudspeaker-IDs always range from 1 to @a number.
 * @param number number of loudspeakers
 * @return @b true on success.
 **/
bool
NewVBAPRenderer::_make_outputs()
{
  // get loudspeakers
  Loudspeaker::container_t loudspeakers;
  _scene.get_loudspeakers(loudspeakers);

  Loudspeaker::container_t::const_iterator i;
  Scene::loudspeakers_t::size_type id = 1;

  /////////// determine loudspeaker delays ////////////////
  for (i = loudspeakers.begin(); i != loudspeakers.end(); ++i)
  {
    DelayLine<sample_t>::size_type delay_samples
      = static_cast<DelayLine<sample_t>::size_type>(
          i->delay * this->get_sample_rate() + 0.5f);

    parameter_map params;
    params["id"] = A2S(id);
    params["delay_line_size"] = A2S(delay_samples + this->get_buffer_size());
    params["delay"] = A2S(delay_samples);
    params["port_name"] = "out_" + A2S(id);

    addOutput(params, *i);

    ++id;
  }

  // Now create the output lists distinguishing subwoofers.
  const output_list_t &outs = get_output_list();
  for (output_list_t::const_iterator i = outs.begin(); i != outs.end(); i++)
  {
    // cull normal loudspeakers 
    if ((*i)->_speaker.model != Loudspeaker::subwoofer)
    {
      _normal_ls_list.push_back(*i);
    }
    // cull subwoofers
    else //if (_output_map.find(i + 1)->second->model == Loudspeaker::subwoofer)
    {
      _subwoofers_list.push_back(*i);
    }
  }

  // sort loudspeakers according to azimuth
  _normal_ls_list.sort(VBAPOutput::sort_operator);

  VERBOSE("NewVBAPRenderer: " << id-1 << " Output objects registered.");

  return true;
}

/** fill a buffer with zeroes.
 * @param buffer pointer to the memory to be cleared
 * @param buffer_size number of data elements to be cleared
 **/
void
NewVBAPRenderer::clear_buffer(sample_t* const buffer,
    const jack_nframes_t buffer_size) const
{
  if (buffer_size != _silence.size())
  {
    // this should never happen because JACK buffer size will not change!
    // XXX: heh ? not true. jack_buffer_size_callback !!!
    // however filling with 0 is faster than copying !!!
    WARNING("BUG?: silence resized! (" << _silence.size() << " -> "
        << buffer_size << ")");
    _silence.resize(buffer_size, 0);
  }
  else
  {
    std::copy(_silence.begin(), _silence.end(), buffer);
  }
}

/** Calculate loudspeaker weights.
 * This function calculates the speaker weight according the the vector base
 * formulation proposed by Ville Pulkki in "Virtual Sound Source Positioning
 * Using Vector Base Amplitude Panning", Journal of the Audio Engineering
 * Society (JAES), Vol.45(6), June 1997.\par
 * The speaker weights are calculated as follows:
 * \f$ \displaystyle g_{left} = \frac{\cos \phi \sin \phi_0 + 
 * \sin \phi \cos \phi_0}{2 \cos \phi_0 \sin \phi_0} \f$
 * \f$ \displaystyle g_{right} = \frac{\cos \phi \sin \phi_0 - 
 * \sin \phi \cos \phi_0}{2 \cos \phi_0 \sin \phi_0} \f$
 * \par 
 * whereby \f$ \displaystyle \phi\f$ denotes blah, blah
 */
NewVBAPRenderer::SpeakerWeights
NewVBAPRenderer::_calculate_speaker_weights(const RenderSource *source
    , const Loudspeaker& left, const Loudspeaker& right)
{
  float phi_0 = (
      ( left.position - _scene.get_reference().position).orientation().azimuth
    - (right.position - _scene.get_reference().position).orientation().azimuth);

  phi_0 = mathtools::wrap_angle(phi_0);

  phi_0 /= 2.0f;

  float phi = (source->src.position
      - _scene.get_reference().position).orientation().azimuth
    - (right.position - _scene.get_reference().position).orientation().azimuth;

  phi = mathtools::wrap_angle(phi);

  phi -= phi_0;

  // TODO: This is an emergency brake.
  if (phi > phi_0) phi = phi_0;
  else if (phi < -phi_0) phi = -phi_0;

  // make sure that sound source lies between two speakers
  assert(phi_0 >= std::abs(phi));

  SpeakerWeights weights;
  if (phi_0 >= 180.f) return weights; // TODO: reject rendering???

  float num1 = cos(deg2rad(phi)) * sin(deg2rad(phi_0));
  float num2 = sin(deg2rad(phi)) * cos(deg2rad(phi_0));
  float den = 2 * cos(deg2rad(phi_0)) * sin(deg2rad(phi_0));

  weights.left  = (num1 + num2) / den;
  weights.right = (num1 - num2) / den;

  // WARNING("g_l: " << weights.left << " g_r: " << weights.right);

  return weights;
}

/** Determines if an angle is larger than another, i.e. if an angle
 * is farther rotated counterclockwise than another
 * @param angle1 angle to be checked against 
 * @param agnle2 reference angle
 * @return @b true if @ angle1 is larger
 */
bool
NewVBAPRenderer::_is_angle_larger(float angle1, float angle2)
{
  return mathtools::wrap_angle(angle1) > mathtools::wrap_angle(angle2);
}

/** Determines if a loudspeaker pair is suitable for rendering a source
 * via VBAP, i.e. if the angle between the loudspeakers is 
 * smaller than 180 degrees
 * @param left loudspeaker left of source
 * @param right loudspeaker right of source
 * @return @b true if loudspeaker pair is suitable
 */
bool
NewVBAPRenderer::_is_valid_loudspeaker_pair(const Loudspeaker& left
    , const Loudspeaker& right)
{
  float phi_0 = (left.position
      - _scene.get_reference().position).orientation().azimuth
    - (right.position - _scene.get_reference().position).orientation().azimuth;

  phi_0 = mathtools::wrap_angle(phi_0);

  return phi_0 < 180.0f;
}

#if 0
/** JACK process callback function.
 * This function is called every time the JACK server has new audio data.
 * @param nframes number of frames
 * @return I don't know what
 */
int
NewVBAPRenderer::jack_process(const jack_nframes_t nframes)
{
// code must be suitable for real-time execution.
// That means following function are not allowed:
// malloc, free, printf, pthread_mutex_lock, sleep,
// wait, poll, select, pthread_join, pthread_cond_wait,
// etc, etc.

  // clear all information about active/inactive loudspeakers
  //_controller.clear_all_active_sources();
  // send some information to the controller (and then to the subscribers)
  //_controller.set_transport_state(this->get_transport_state());
  //_controller.set_cpu_load(this->get_cpu_load(), 0);

      {
        Loudspeaker::container_t all_loudspeakers;
        // necessary memory is reserved inside the function
        _scene.get_loudspeakers(all_loudspeakers);
        // the loudspeakers contain already absolute positions

        // TODO: remove this if clause?
        if (all_loudspeakers.size() != _output_map.size())
        {
          WARNING("Different number of loudspeakers in Scene and NewVBAPRenderer!");
          // TODO: how to react on that?
        }

        output_list_t normal_ls_list;
        output_list_t subwoofers_list;



        const DirectionalPoint reference = _scene.get_reference();
	// float master_volume = _scene.get_corrected_master_volume();

	// iterate over all sources
        for (input_map_t::const_iterator in_iter = _input_map.begin();
             in_iter != _input_map.end(); ++in_iter)
        {
          assert(in_iter->second != NULL);

          // determine active loudspeakers for each source
          output_list_t::iterator iter_to_loudspeaker = normal_ls_list.begin();
          output_list_t::iterator iter_to_next_loudspeaker = ++normal_ls_list.begin();

          // if only one loudspeaker present
          if (iter_to_next_loudspeaker == normal_ls_list.end())
          {

	    _render_source_on_output(*in_iter->second, *iter_to_loudspeaker, 
				     1.0f, nframes);

            continue; // next source
          }

          bool ls_pair_found = false;

          // iterate to loudspeaker left of source
          while (!ls_pair_found && iter_to_next_loudspeaker != normal_ls_list.end())
          {
            // check if we have surpassed the source
            if (_is_angle_larger((((*iter_to_loudspeaker)->position
                 - _scene.get_reference().position).orientation()
            - (*_scene.get_source_position(in_iter->first)
               - _scene.get_reference().position).orientation()).azimuth,
               (((*iter_to_next_loudspeaker)->position
                 - _scene.get_reference().position).orientation()
            - (*_scene.get_source_position(in_iter->first)
               - _scene.get_reference().position).orientation()).azimuth))
            {
              ls_pair_found = true;
            }
            else
            {
              iter_to_loudspeaker++;
              iter_to_next_loudspeaker++;
            }

          } // while

          // if between first and last
          if (!ls_pair_found && iter_to_next_loudspeaker == normal_ls_list.end())
          {
            iter_to_loudspeaker = --normal_ls_list.end();
            iter_to_next_loudspeaker = normal_ls_list.begin();
          }

          if (!_is_valid_loudspeaker_pair(**iter_to_next_loudspeaker,
                  **iter_to_loudspeaker))
          {
            // find loudspeaker which is closer
            float angle_to_left_loudspeaker = (((*iter_to_next_loudspeaker)->position
                  - _scene.get_reference().position).orientation()
                - (*_scene.get_source_position(in_iter->first) 
                  - _scene.get_reference().position).orientation()).azimuth;

            if (angle_to_left_loudspeaker > 180.0f) 
            {
              angle_to_left_loudspeaker = std::abs(angle_to_left_loudspeaker - 360);
            }

            float angle_to_right_loudspeaker = (((*iter_to_loudspeaker)->position
                  - _scene.get_reference().position).orientation()
                - (*_scene.get_source_position(in_iter->first)
                  - _scene.get_reference().position).orientation()).azimuth;

            if (angle_to_right_loudspeaker > 180.0f)
            {
              angle_to_right_loudspeaker = std::abs(angle_to_right_loudspeaker - 360);
            }

            Output* active_loudspeaker = NULL;

            float weight = 1.0f;

            if (_is_angle_larger(angle_to_right_loudspeaker, angle_to_left_loudspeaker))
            {
              if (angle_to_left_loudspeaker < 30.0f)
              {
                active_loudspeaker = *iter_to_next_loudspeaker;
                weight = 0.5f * (1 + cos(mathtools::PI/30.0f * angle_to_left_loudspeaker));
              }
            }
            else
            {
              if (angle_to_right_loudspeaker < 30.0f)
              {
                active_loudspeaker = *iter_to_loudspeaker;
                weight = 0.5f * (1 + cos(mathtools::PI/30.0f * angle_to_right_loudspeaker));
              }
            }

            if (active_loudspeaker != NULL)
            {
              VERBOSE2("Couldn't find valid loudspeaker pair for source '"
                  << _scene.get_source_name(in_iter->first) << "'. Using closest loudspeaker only.");

              // render source
	      _render_source_on_output(*in_iter->second, active_loudspeaker, 
				       weight, nframes);
              // render subwoofers
              for (output_list_t::const_iterator i = subwoofers_list.begin();
                  i != subwoofers_list.end(); ++i)
              {
		_render_source_on_output(*in_iter->second, *i, weight, nframes);
              }
            }
            else
            {
              VERBOSE2("Couldn't find valid loudspeaker pair for source '"
                  << _scene.get_source_name(in_iter->first) << "'.");
            }

            continue; // next source
          }
          else
          {
            SpeakerWeights weights = _calculate_speaker_weights(*in_iter->second,
                      **iter_to_next_loudspeaker,
                      **iter_to_loudspeaker);

            // render source
	    _render_source_on_output(*in_iter->second, *iter_to_next_loudspeaker, 
				     weights.left, nframes);
	    
	    _render_source_on_output(*in_iter->second, *iter_to_loudspeaker, 
				     weights.right, nframes);

            // render subwoofers
            for (output_list_t::const_iterator i = subwoofers_list.begin();
                i != subwoofers_list.end(); ++i)
            {
	      _render_source_on_output(*in_iter->second, *i, 1.0f, nframes);
            }
          }
        } // for all sources

        float master_level = 0.0f;

	// finish rendering on all outputs
        for (output_map_t::const_iterator i = _output_map.begin();
             i != _output_map.end(); ++i)
        {
          assert(i->second != NULL);
	  
	  // fade-out sources on loudspeakers which are not used anymore
	  while (i->second->unrendered_source_left())
	  {
	    _fade_out_source_on_output(
			    i->second->return_id_of_remaining_source(), 
			    &(*i->second), nframes);
	  }

	  // finally release audio data
	  i->second->write_output_buffer(nframes);

          master_level = std::max(master_level, i->second->get_signal_level(nframes));

          if (i->second->active_source_ids.size() != 0)
          {
            _controller.set_active_sources(i->first,
                i->second->active_source_ids);
          }
        }

        _controller.set_master_signal_level(master_level, 0);

        // return something?
        break;
      }
  return 0;
}
#endif

void
NewVBAPRenderer::_render_source_on_output(const RenderSource* render_source
    , Output* output, float weighting_factor, const jack_nframes_t nframes)
{
  const Source &source = render_source->src;
  VBAPOutput *current_output = dynamic_cast<VBAPOutput *>(output);
  VERBOSE3("\n======================================"
      "==========================================");
  VERBOSE3("== output position: x: " << current_output->_speaker.position.x
      << ", y: " << current_output->_speaker.position.y
      << ", azimuth: " << current_output->_speaker.orientation.azimuth);
  //VERBOSE3("== JACK port: " << current_output->get_port_name());

  // no distance attenuation for plane waves 
  if (source.model == Source::plane)
  {
    weighting_factor *= 0.5f / _scene.get_amplitude_reference_distance(); // 1/r
    //weighting_factor *= 
    //     0.25f / sqrt(_scene.get_amplitude_reference_distance()); // 1/sqrt(r)
  }
  else
  {
    // consider distance attenuation
    float source_distance = (source.position
        - _scene.get_reference().position).length();

    // no volume increase for sources closer than 0.5m
    source_distance = std::max(source_distance, 0.5f);
    weighting_factor *= 0.5f / source_distance; // 1/r
    //weighting_factor *= 0.25f / sqrt(source_distance); // 1/sqrt(r)
  }

  // apply the gain factor of the current source
  weighting_factor *= source.gain;

  // apply master volume
  weighting_factor *= _scene.get_corrected_master_volume();

  // apply loudspeaker weight
  weighting_factor *= current_output->_speaker.weight;

  if (source.mute == true) weighting_factor = 0.0f;

  assert(weighting_factor >= 0.0f);

  // input object must not be deleted! It's deleted in ~NewVBAPRenderer()
  const VBAPInput* current_input
    = dynamic_cast<VBAPInput *>(render_source->get_input());
  if (current_input == NULL)
  { // this happens regularly when a source is added and the renderer is
    // already processing because it takes some time until the JACK ports are
    // ready. Therefore the following warning is only printed if verbose > 1.
    VERBOSE2("NewVBAPRenderer::Output: Input for source " << render_source->id
       << " doesn't exist in input map!");
    return; //continue; // continue with next source
  }

  float previous_source_weight
    = current_output->get_previous_source_weight(render_source->id);

  const sample_t* in = current_input->get_sample_data();
  sample_t* out = current_output->get_output_buffer(nframes);

  // this is the actual audio processing
  weight_buffer_and_add(in, previous_source_weight, weighting_factor
      , out, nframes);

  current_output->store_current_source_weight(render_source->id
      , weighting_factor);

  // The current loudspeaker is active for the current source:
  current_output->active_source_ids.insert(render_source->id);
}

void
NewVBAPRenderer::_fade_out_source_on_output(const RenderSource *render_source
    , Output* output, const jack_nframes_t nframes)
{
  const VBAPInput* current_input
    = dynamic_cast<VBAPInput *>(render_source->get_input());
  VBAPOutput* current_output = dynamic_cast<VBAPOutput *>(output);

  if (current_input == NULL)
  {
    VERBOSE2("NewVBAPRenderer::Output: Input for source " << render_source->id
        << " doesn't exist in input map!");
    return;
  }

  const sample_t* in = current_input->get_sample_data();
  sample_t* out = current_output->get_output_buffer(nframes);

  float previous_source_weight
    = current_output->get_previous_source_weight(render_source->id);

  // this is the actual audio processing
  weight_buffer_and_add(in, previous_source_weight, 0.0f, out, nframes);
}

void
NewVBAPRenderer::weight_buffer_and_add(const sample_t* buffer_to_be_added
    , const float prev_weight, const float curr_weight, sample_t* target_buffer
    , const jack_nframes_t nframes)
{
  // weight audio data and add to output buffer  
  for (unsigned int sample = 0; sample < nframes; sample++)
  {
    // TODO: Make crossfade more elegant?
    *(target_buffer+sample) += *(buffer_to_be_added+sample) *
      (curr_weight * _fade_in[sample] + prev_weight * _fade_out[sample]);
  }
}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class NewVBAPRenderer::Output
////////////////////////////////////////////////////////////////////////////////

/** ctor.
 * @param renderer calling NewVBAPRenderer object
 **/
NewVBAPRenderer::VBAPOutput::VBAPOutput(NewRenderer& renderer
    , const parameter_map &params)
  : NewRenderer::Output(renderer, params)
  , DelayLine<sample_t>(params.get<int>("delay_line_size")
      , params.get<int>("delay"))
  , _delay_line_required(params.get<int>("delay"))
{
  _audio_buffer.resize(_renderer.get_buffer_size(), 0.0f);
}

NewVBAPRenderer::VBAPOutput::~VBAPOutput()
{}

void
NewVBAPRenderer::VBAPOutput::cycle_start()
{
  update_previous_source_weights();
  clear_active_source_ids();

  // clear output buffer (to accumulate data)
  clear_output_buffer(_renderer._nframes);
}

float*
NewVBAPRenderer::VBAPOutput::get_output_buffer(const jack_nframes_t nframes)
{
  (void) nframes;
  if (!_delay_line_required)
  {
    // return JACK output buffer
    return get_sample_destination();
  }
  // return internal audio buffer
  else return &_audio_buffer[0];
}

void
NewVBAPRenderer::VBAPOutput::write_output_buffer(const jack_nframes_t nframes)
{
  if (!_delay_line_required) return; // nothing to do

  // else: write to delay line, then read from it

  this->prepare_write();
  this->write(&_audio_buffer[0], nframes);

  // get JACK output buffer
  sample_t* out = static_cast<sample_t*>(this->get_sample_destination());

  if (!read(out, nframes))
  {
    ERROR("Can not write output buffer.");
  }
}

void
NewVBAPRenderer::VBAPOutput::update_previous_source_weights()
{
  _previous_source_weights.clear();
  _previous_source_weights = _current_source_weights;
  _current_source_weights.clear();
}

float
NewVBAPRenderer::VBAPOutput::get_previous_source_weight(id_t source_id)
{
  coefficient_map_t::iterator iter
    = _previous_source_weights.find(source_id);
  float weight = 0.0f;

  if (iter != _previous_source_weights.end())
  {
    weight = iter->second.weight;
    _previous_source_weights.erase(iter);
  }
  return weight;
}

void
NewVBAPRenderer::VBAPOutput::store_current_source_weight(const id_t id
    , const float weight)
{
  _current_source_weights[id] = Coefficient(weight);
}

bool
NewVBAPRenderer::VBAPOutput::unrendered_source_left() const
{
  return !_previous_source_weights.empty();
}

id_t
NewVBAPRenderer::VBAPOutput::return_id_of_remaining_source()
{
  id_t id = _previous_source_weights.begin()->first;

  // erase source
  _previous_source_weights.erase(_previous_source_weights.begin());
  return id;
}

void
NewVBAPRenderer::VBAPOutput::write_empty_buffer(const jack_nframes_t nframes)
{
  NewVBAPRenderer *renderer = dynamic_cast<NewVBAPRenderer *>(&_renderer);
  sample_t *out = get_sample_destination();
  renderer->clear_buffer(out, nframes);
}

void
NewVBAPRenderer::VBAPOutput::clear_active_source_ids()
{
  this->active_source_ids.clear(); // reset active sources to an empty set
}

void
NewVBAPRenderer::VBAPOutput::clear_output_buffer(const jack_nframes_t nframes)
{
  NewVBAPRenderer *renderer = dynamic_cast<NewVBAPRenderer *>(&_renderer);
  sample_t* out;

  if (!_delay_line_required)
  {
    out = static_cast<sample_t*>(this->get_sample_destination());
  }
  else
  {
    out = &_audio_buffer[0];
  }
  renderer->clear_buffer(out, nframes);
}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class NewVBAPRenderer::Source
////////////////////////////////////////////////////////////////////////////////

NewVBAPRenderer::VBAPSource::VBAPSource(NewRenderer &renderer, Input *input
    , const parameter_map &params)
  : NewRenderer::RenderSource(renderer, input, params)
{}

void
NewVBAPRenderer::VBAPSource::cycle_start()
{
  find_speaker_pair();
}

void
NewVBAPRenderer::VBAPSource::find_speaker_pair()
{
  NewVBAPRenderer *renderer = dynamic_cast<NewVBAPRenderer *>(&_renderer);
  output_list_t &ls_list = renderer->_normal_ls_list;
  const DirectionalPoint reference = renderer->_scene.get_reference();
  const Position refpos = reference.position;
  const Position srcpos = src.position;

  // determine active loudspeakers for each source
  output_list_t::iterator iter_to_loudspeaker = ls_list.begin();
  output_list_t::iterator iter_to_next_loudspeaker = ++(ls_list.begin());

  // if only one loudspeaker present
  if (iter_to_next_loudspeaker == ls_list.end())
  {
    left_speaker = *iter_to_loudspeaker;
    right_speaker =  NULL;
    weight_left = 1.0f;
    weight_right = 0.0f;
    return;
  }

  bool ls_pair_found = false;

  // iterate to loudspeaker left of source
  while (!ls_pair_found && iter_to_next_loudspeaker != ls_list.end())
  {
    // check if we have surpassed the source
    if (renderer->_is_angle_larger(((
              (*iter_to_loudspeaker)->_speaker.position - refpos).orientation()
            - (srcpos - refpos).orientation()).azimuth, ((
              (*iter_to_next_loudspeaker)->_speaker.position - refpos).orientation()
            - (srcpos - refpos).orientation()).azimuth))
    {
      ls_pair_found = true;
    }
    else
    {
      iter_to_loudspeaker++;
      iter_to_next_loudspeaker++;
    }

  } // while

  // if between first and last
  if (!ls_pair_found && iter_to_next_loudspeaker == ls_list.end())
  {
    iter_to_loudspeaker = --ls_list.end();
    iter_to_next_loudspeaker = ls_list.begin();
  }

  if (!renderer->_is_valid_loudspeaker_pair(
        (*iter_to_next_loudspeaker)->_speaker,(*iter_to_loudspeaker)->_speaker))
  {
    // find loudspeaker which is closer
    float angle_to_left_loudspeaker
      = (((*iter_to_next_loudspeaker)->_speaker.position
            - refpos).orientation() - (srcpos - refpos).orientation()).azimuth;

    if (angle_to_left_loudspeaker > 180.0f)
    {
      angle_to_left_loudspeaker = std::abs(angle_to_left_loudspeaker - 360);
    }

    float angle_to_right_loudspeaker
      = (((*iter_to_loudspeaker)->_speaker.position - refpos).orientation()
        - (srcpos - refpos).orientation()).azimuth;

    if (angle_to_right_loudspeaker > 180.0f)
    {
      angle_to_right_loudspeaker = std::abs(angle_to_right_loudspeaker - 360);
    }

    left_speaker = NULL;
    right_speaker = NULL;

    weight_left = 1.0f;
    weight_right = 0.0f;

    if (renderer->_is_angle_larger(angle_to_right_loudspeaker
          , angle_to_left_loudspeaker))
    {
      if (angle_to_left_loudspeaker < 30.0f)
      {
        left_speaker = *iter_to_next_loudspeaker;
        weight_left = 0.5f * (1 + cos(mathtools::PI/30.0f
              * angle_to_left_loudspeaker));
      }
    }
    else
    {
      if (angle_to_right_loudspeaker < 30.0f)
      {
        left_speaker = *iter_to_loudspeaker;
        weight_left = 0.5f * (1 + cos(mathtools::PI/30.0f
              * angle_to_right_loudspeaker));
      }
    }

    if (left_speaker != NULL)
    {
      VERBOSE2("Couldn't find valid loudspeaker pair for source '"
          << this->id << "'. Using closest loudspeaker only.");
    }
    else
    {
      VERBOSE2("Couldn't find valid loudspeaker pair for source '"
          << this->id << "'.");
    }
  }
  else
  {
    SpeakerWeights weights = renderer->_calculate_speaker_weights(this,
        (*iter_to_next_loudspeaker)->_speaker,
        (*iter_to_loudspeaker)->_speaker);

    left_speaker = *iter_to_next_loudspeaker;
    right_speaker = *iter_to_loudspeaker;
    weight_left = weights.left;
    weight_right = weights.right;
  }
}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class NewVBAPRenderer::Input
////////////////////////////////////////////////////////////////////////////////

// XXX: this is basically useless, the full fledged NewRenderer::Input will
// provide all the required functionality.
// i just leave it here, to have this code around.

/** private ctor. Object can only be created with create().
 * @param renderer reference to the calling NewVBAPRenderer object. Using this,
 * we can access the public members of the VBAP renderer.
 **/
NewVBAPRenderer::VBAPInput::VBAPInput(NewRenderer& renderer
    , const parameter_map &params)
  : NewRenderer::Input(renderer, params)
{}

NewVBAPRenderer::VBAPInput::~VBAPInput()
{}

void
NewVBAPRenderer::VBAPInput::cycle_start()
{
  // TODO: refactor fetch_new_buffer to here.
  NewRenderer::Input::cycle_start(); // call superclass for metering.
}

bool
NewVBAPRenderer::VBAPInput::fetch_new_buffer(jack_nframes_t nframes)
{
  (void) nframes;
  const sample_t *in;
  in = static_cast<const sample_t*>(get_sample_data());
  if (in == NULL)
  {
    WARNING("Got no input buffer from JACK!");
    return false;
  }
  assert(in != NULL);

#if 0
  if (_renderer._scene.get_source_mute_state(this->id))
  {
    // update source signal level
    //_controller.set_source_signal_level(this->id, 0.0f, 0);
  }
  else
  {
    // update source signal level (max amplitude)
    //_controller.set_source_signal_level(this->id,
    //    _renderer.get_max_signal_amplitude(_current_audio_frame, nframes)
    //	     * _scene.get_source_gain(this->id), 0);

    // update source signal level (RMS)
    //_controller.set_source_signal_level(_source_id, 
    //      _renderer.get_signal_rms(in, nframes) * 
    //      std::pow(_scene.get_source_gain(_source_id), int(nframes))
    //      , 0);
  }
#endif

  return true;
}

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
