/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * Abstract NewRenderer and Renderer classes (definition). 
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#ifndef RENDERER_H
#define RENDERER_H

#include <semaphore.h>
#include <pthread.h>

#include <list>   // for port_list_t
#include <string> //    -- " --

#include "jackclient.h" // base class
#include "scene.h"

class Controller;

#include "rtcommand.h"
#include "parameter_map.h"
#include "classtools.h" // for NonCopyable

// A Map which holds parameters while constructing different
// SubClasses of Renderers and Inputs.
// Constructors will pick relevant fields.
//typedef std::map<std::string,std::string> parameter_map_t;

class Lockable : classtools::NonCopyable
{
  public:
    Lockable()
    {
      if (pthread_mutex_init(&_lock, NULL))
      {
        throw std::runtime_error("Can't init mutex. (impossible !!!)");
      }
    }

    int   lock() { return pthread_mutex_lock(  &_lock); }
    int unlock() { return pthread_mutex_unlock(&_lock); }

    class ScopedLock : classtools::NonCopyable
    {
      public:
        explicit ScopedLock(Lockable* obj) : _obj(obj) { _obj->lock(); }
        ~ScopedLock() { _obj->unlock(); }

      private:
        Lockable* const _obj;
    };

  private:
    pthread_mutex_t _lock;
};

/** Parallel rendering engine.  
 * This is the base class for parallel renderers. It encapsulates parallel 
 * computation from the implementation of the various renderers. 
 *
 * Three different "basic-operation-classes" are encapsulated as nested classes
 * in NewRenderer: 
 * - Input
 * - Output 
 * - RenderSource
 *
 * Each parallel renderer, e.g. NewBinauralRenderer, implements NewRenderer and 
 * its nested classes. NewRenderer holds "job-lists" containing the 
 * instances of the basic-operation-classes, e.g. NewRenderer::_input_list. 
 *
 * For each basic-operation-class audio computation is divided into in three 
 * "stages": 
 * - cycle_start: Compute jobs before rendering stage, e.g. metering input. 
 * - render: Render and write audio data to a JACK port.   
 * - cycle_end: Compute jobs after rendering stage, e.g. write an empty buffer 
 *   to the JACK port if rendering was not computed.  
 *
 * This virtual functions are declared in the basic-operation-classes and can
 * be implemented in child classes. The child classes are the renderers which
 * are not forced to implement all stages. 
 * 
 * The function do_in_this_thread() will iterate over the job lists in each 
 * stage and execute the implementation of the stages in the child classes via 
 * late binding. 
 *
 * @see NewWFSRenderer
 * @see NewBinauralRenderer 
 * @see NewVBAPRenderer
 *
 * Synchronization of parallel computation is achieved via this rules: 
 * - The function jack_process() triggers all "slave audio threads". It is 
 *   driven by JACK calling this function at each audio cycle.
 * - Each stage terminates before the next begins synchronized in all threads. 
 *   This is achieved in jack_process() by calling wait_slave_threads() and 
 *   cont_slave_threads(). 
 * - The mapping of jobs to threads is computed in the function 
 *   do_in_this_thread(). 
 *
 * @see jack_process()
 * @see do_in_this_thread()
 * @see slave_audio_thread()
 *
 * The parallel rendering engine uses the non-blocking datastructure RTList to
 * communicate between realtime and non-realtime threads.
 * All non-realtime accesses to RTList's have to be locked with lock()/unlock()
 * or with ScopedLock to ensure single-reader/single-writer operation.
 * See, e.g., addSourceInput().
 **/
class NewRenderer : public JackClient, private Lockable
{
  public:
    /// Ptr to NewRenderer
    typedef std::auto_ptr<NewRenderer> ptr_t;
    /// JACK port list
    typedef std::list<std::string> port_list_t;

    /// Holds parameters about SSR
    const parameter_map params;

    /// Self explanatory
    void get_output_port_names(port_list_t& port_list) const;

    /// Self explanatory
    unsigned int get_number_of_output_channels() const;

  protected:
    class Input;
    class Output;
    class RenderSource;

    /// %Renderer exception  
    struct renderer_error : public std::runtime_error
    {
      renderer_error(const std::string& s): std::runtime_error(s) {}
    };

    NewRenderer(const Scene& scene, Controller& controller,
        const parameter_map &params_) throw (jack_error, renderer_error);

    /** List of Output Objects.
     * @note It is not a RTList.
     **/
    typedef std::list<Output*> output_list_t;

  private:
    // The outputs are static. So a Map of (Output *) is good.
    // This map will get built from the loudspeakers array in the Scene.
    //typedef std::map<Loudspeaker::container_t::size_type, Output*> output_map_t;

    //its not yet totally clear to me, what the implications of iusos Safe Set 
    //are.

    /** RTList of Input. 
     * The Inputs may change and especially need to be synchronised with 
     * Sources.
     * @note Several mirror sources, could share a single input.
     **/
    typedef RTList<Input> input_list_t;
    /// RTList of RenderSource
    typedef RTList<RenderSource> source_list_t;
    /** Non RT map of Source.
     * @note A Source is a member of RenderSource. 
     **/
    typedef std::map<int,Source> shadow_source_map_t;

    /// RTCommandEngine contains the lockfreefifo
    RTCommandEngine _engine;

    /// @name Metering variables
    //@{
    /// Meter semaphore triggered in jack_process().
    sem_t _meter_semaphore;
    /// Set by ctor/dtor. Controlls lifetime of meteringThread().
    volatile bool _exiting;
    /// The thread called
    pthread_t _meter_threadid;
    ///@}

    /** @name Thread variables
     * Variables for threads that must be started and maintained.
     * @see slave_audio_thread()
     * @see jack_process()
     **/
    //@{
    /// Amount of slave threads
    const int _num_threads;
    /// Ids of slaves
    std::vector<pthread_t> _slave_thread_ids;
    /** Vector containing waiting semaphores. 
     * One for each thread.
     * @see wait_slave_threads()
     **/
    std::vector<sem_t> _slave_wait_semaphores;
    /** Vector containing continued semaphores.
     * One for each thread.
     * @see cont_slave_threads()
     **/
    std::vector<sem_t> _slave_cont_semaphores;
    /// @}


    static void *meteringThreadAux(void *r);
    void meteringThread();
    void publishMeter();

    static void * slave_audio_thread_aux(void *data);
    void slave_audio_thread(int thread_number);
    bool do_in_this_thread(int current_job, int thread_number);
    void create_slave_audio_threads(int number);
    void wait_slave_threads();
    void cont_slave_threads();

  protected:
    // Every Renderer defines its own Input and Output class,
    // which are derived from Input and Output.
    /** Create an Input.
     * @param params
     **/
    virtual Input *create_input(const parameter_map &params) = 0;
    /** Create an Output. 
     * @param params
     **/
    virtual Output *create_output(const parameter_map &params) = 0;

    // RenderSource. (i will keep this simple first)
    // I will relate the source to an input. however for mirror
    // sources there would generally be one master source, which 
    // controls other sources. But this might be part of the renderer
    // implementation. 
    // the relation from source to input is not happening here.
    // it might even be implicit first.
    /** Create a RenderSource.
     * @param input An Input
     * @param params
     **/
    virtual RenderSource *create_source(Input *input
        , const parameter_map &params) = 0;

    // The Input associated with a Source is also passed 
    //
    // This will hide the Datastructures containing Ins and Outs
    // from the renderers. 
    // Need to check if any per process_cb initializing is done
    // somewhere. But most could be done using cycle_start etc.
    //
    /** Call render for every (RenderSource, Output) pair.
     * The renderer will dynamic_cast the pointers to its IO class.
     * @param rendersource Ptr. to a RenderSource
     * @param output Ptr to an Output
     **/
    virtual void render(RenderSource *rendersource, Output *output) = 0;

    // XXX: renderer should not be allowed to access unsafe datastructures.
    //      so this might get private.
    const Scene &_scene; ///< Reference to a Scene 
    Controller &_controller; ///< Reference to the Controller

    // addOutput to be called from constructor only.
    // perhaps i should also make the outputs RTSafe.
    //   XXX: no. too much cruft in the renderers.
    //        this needs a lot of sorting out.
    void addOutput(const parameter_map &params);

    // XXX: hmm... this might be good.
    void addOutput(const parameter_map &params, const Loudspeaker &speaker);

  public:
    virtual ~NewRenderer();

    // XXX: Perhaps this should be protected ?
    //      either protected, like now, or we add the 
    //      nframes param to all necessary
    //      functions. which would suck.
    jack_nframes_t _nframes;  ///< From JACK 

    // Generic Renderer Factory. 
    // the parameter map will hold all old constructor parameters.
    // and also the type of the Renderer to create.
    static ptr_t create(const Scene& scene, Controller& controller,
        const parameter_map &params);

    // only add some functions for testing now, until i decide on the 
    // real interface.

    void addSourceInput(const parameter_map &params);
    void remSourceInput(int id);
    void clearAllSourceInput();

    Source getSourceCopy(int id);
    void pushSourceChange(int id, Source s);
    float get_meter(sample_t*);

  private:
    int jack_process(jack_nframes_t nframes);

    input_list_t  _input_list;    ///< All input objects are stored here
    source_list_t _source_list;   ///< All source objects are stored here
    output_list_t _output_list;   ///< All output objects are stored here

    /// Non RT map of Source.
    shadow_source_map_t _shadow_sources;

    const std::string   _input_prefix;  ///< Prefix for the input ports
    const std::string   _output_prefix; ///< Prefix for the output ports

  public:
    /// Return list of Outputs. 
    const output_list_t& get_output_list() const { return _output_list; }
};

/** An audio Input.
 * For now it is a jack port or a FileStream.
 *
 * So i need to provide a Stream Class, where the distinction
 * between JackPort and FileStream will happen. But its all
 * private, so i go with a JackPort only first.
 **/
class NewRenderer::Input : classtools::NonCopyable
{
  protected:
    NewRenderer& _renderer; ///< Which renderer. 
    Input(NewRenderer &renderer, const parameter_map &params);

  private:
    jack_port_t* _port; ///< A JACK port. 

  public:
    float *get_sample_data() const;
    /** Start a cycle.
     * Happens before rendering.
     **/
    virtual void cycle_start() {}
    /** End a cycle.
     * Happens after rendering.
     **/
    virtual void cycle_end() {}
    virtual ~Input();

    const std::string port_name; ///< full name of the JACK port
};

/** Consists of a Source and an Input.   
 * Makes an arbitrary Input be a source of sound by connecting the
 * audio signal with the Scene. 
 **/
class NewRenderer::RenderSource : classtools::NonCopyable
{
  private:
    std::map<int,bool> _activity_map; ///< Under development  

    float _meter_max; ///< Maximum level of RenderSource. 
    //float _meter_rms; ///< Average (RMS) level of RenderSource. 
  protected:
    NewRenderer &_renderer; ///< Which NewRenderer
    Input *_input; ///< Pointer to an Input. 
    RenderSource(NewRenderer &renderer, Input *input,
        const parameter_map &params);

  public:
    /** Start a cycle.
     * Happens before rendering.
     **/
    virtual void cycle_start() {}
    /** End a cycle.
     * Happens after rendering.
     **/
    virtual void cycle_end() {}
    virtual ~RenderSource() {}

    void set_output_activity(int output, bool active);
    void meter();

    /// Returns the maximum level.    
    float get_meter_level() const { return _meter_max; }
    bool output_active(int output);

    // XXX: this is tmp.
    /// Returns the Input of this RenderSource. 
    Input *get_input() const { return _input; }
    int id; ///< ID of RenderSource.
    Source src; ///< Some (wave) Source. 
};

/// An audio Output. 
class NewRenderer::Output : classtools::NonCopyable
{
  private:
    static float _master_meter_max; ///< Maximum level of all outputs. 
    float _channel_meter_max; ///< Maximum level of one Output channel. 
  protected:
    NewRenderer &_renderer; ///< Which renderer. 
    Output(NewRenderer &renderer, const parameter_map &params);

  private:
    jack_port_t* _port; ///< A JACK port. 

  public:
    int id; ///< ID of Output. 
    Loudspeaker _speaker; ///< A Loudspeaker obj. 
    sample_t *get_sample_destination();
    const std::string port_name; ///< full name of the JACK port

    /** Retruns the maximum level.
     * @see meter_master()
     * @return _master_max  
     **/
    float get_meter_level() const {return _master_meter_max;}

    /** Start a cycle.
     * Happens before rendering.
     **/
    virtual void cycle_start() {}
    /** End a cycle.
     * Happens after rendering.
     **/
    virtual void cycle_end() {}
    void meter_master();
    virtual ~Output();
};

// Not sure if this is the best place.
// An RTCommand which fetches all meter levels, and publishes them
// in the cleanup stage.

// Now the old code follows. this stuff remains, to not totally
// break compiling.

/** Abstract renderer class.
 * Renderers like WFSRenderer and BinauralRenderer can be derived from this base
 * class. This class manages the communication with the main program, any audio
 * related tasks are done in the derived classes.
 **/
class Renderer : public JackClient
{
  public:
    typedef std::list<std::string> port_list_t; ///< a list of JACK port names

    /// Type of renderer
    typedef enum
    {
      WFS,       ///< Wave Field Synthesis
      binaural,  ///< well, binaural
      generic,   ///< generic renderer
      BRS,       ///< Binaural Room Scanning
      BPB,       ///< Binaural playback
      VBAP,      ///< Vector-Base Amplitude Panning
      AAP,       ///< traditional Ambisonics amplitude panning
      Ambisonics ///< Ambisonics (not implemented)
    } renderer_t;

    /// dtor. 
    virtual ~Renderer() {}

    /** Add a new source to the rendering setup. 
     * That can include reserving memory, assigning and connecting ports and 
     * such things.
     * @param id unique identifier of the input channel
     * @param port_name name of JACK port associated with the new source.
     * @return @a true on success
     * @see delete_source
     **/
    virtual bool new_source(const id_t id, const std::string& port_name) = 0;

    /// BRS renderer needs this. 
    // TODO: Do it more elegant!!!
    virtual bool new_source(const id_t id, const std::string& port_name,
        const std::string& properties_file) = 0;

    /** Delete a source from the rendering setup. 
     * That can include freeing memory, unassigning ports and such things.
     * @param id unique identifier of the source
     * @return @a true on success
     * @see new_source
     **/
    virtual bool delete_source(const id_t id) = 0;

    /** Delete all source from the rendering setup. 
     * That can include freeing memory, unassigning ports and such things.
     **/
    virtual void delete_all_sources() = 0;

    /// Get names of all output ports.
    virtual void get_output_port_names(port_list_t& port_list) const = 0;

    /// Self explanatory
    virtual unsigned int get_number_of_output_channels() const = 0;

    /// Returns what type the current renderer actually is.
    virtual renderer_t get_renderer_type() const = 0;

  protected:
    /** ctor. 
     * @param scene Reference to the scene used by the calling class
     * @param client_name Name of Client
     * @param use_jack_process_callback 
     **/
    Renderer(const Scene& scene, const std::string& client_name,
       bool use_jack_process_callback = false) throw (jack_error) :
      JackClient(client_name, use_jack_process_callback), // base class ctor.
      _scene(scene)
    {}

    /// Reference to the scene used in the calling class
    const Scene& _scene;
};

#endif

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
