/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * %Scene class (implementation).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#include <cassert>

#include "scene.h"
#include "source.h"

Scene::Scene() :
  // reference looking straight ahead (in positive y-direction)
  _reference(Position(0, 0), Orientation(90)),
  _master_volume(1.0f), _master_volume_correction(1.0f), 
  _amplitude_reference_distance(3.0f),
  _master_signal_level(0.0f),
  _cpu_load(0.0f), _sample_rate(0u)
{}

Scene::~Scene()
{
  //maptools::purge(_source_map);
}

Scene::loudspeakers_t::size_type Scene::get_number_of_loudspeakers() const
{
  return _loudspeakers.size();
}

void Scene::set_loudspeakers(const Loudspeaker::container_t& loudspeakers)
{
  // mind the difference between _loudspeakers and loudspeakers!
  if (_loudspeakers.size() != 0)
  {
    ERROR("BUG: Loudspeaker list can only be set once for a Scene!");
    _loudspeakers.clear();
  }
  // reserve memory to avoid unnecessary re-allocation
  _loudspeakers.reserve(loudspeakers.size());
  for (Loudspeaker::container_t::const_iterator i = loudspeakers.begin();
      i != loudspeakers.end(); ++i)
  {
    // construction of temporary variable with type conversion ctor
    // and copy construction into vector
    _loudspeakers.push_back(PrivateLoudspeaker(*i));
  }
}

void Scene::new_source(const id_t id)
{
  // do nothing if id already exists:
  if (maptools::get_item(_source_map, id) == false)
  {
     VERBOSE("Adding source " << id << " to source map!");
    _source_map[id] = Source(); // first standard ctor., then assignment.
  }
}

void Scene::delete_source(id_t id)
{
  // this should call the destructor for the Source object.
  // IMPORTANT: the map holds the Sources directly, no pointers!
  _source_map.erase(id);
}

void Scene::delete_all_sources()
{
  // this should call the destructor for all Source objects.
  // IMPORTANT: the map holds the Sources directly, no pointers!
  _source_map.clear();
}

bool Scene::set_source_position(id_t id, const Position& position)
{
  return _set_source_member(id, &Source::position, position);
}

bool Scene::set_source_orientation(id_t id, const Orientation& orientation)
{
  return _set_source_member(id, &Source::orientation, orientation);
}

bool Scene::set_source_gain(id_t id, const float& gain)
{
  return _set_source_member(id, &Source::gain, gain);
}

bool Scene::set_source_signal_level(id_t id, const float& level)
{
  return _set_source_member(id, &Source::signal_level, level);
}


bool Scene::set_source_mute(id_t id, const bool& mute)
{
  return _set_source_member(id, &Source::mute, mute);
}

bool Scene::set_source_name(id_t id, const std::string& name)
{
  return _set_source_member(id, &Source::name, name);
}

bool Scene::set_source_properties_file(id_t id, const std::string& name)
{
  return _set_source_member(id, &Source::properties_file, name);
}

bool Scene::set_source_port_name(id_t id, const std::string& port_name)
{
  return _set_source_member(id, &Source::port_name, port_name);
}

bool Scene::set_source_file_name(id_t id, const std::string& file_name)
{
  return _set_source_member(id, &Source::audio_file_name, file_name);
}

bool Scene::set_source_file_channel(id_t id, const unsigned int& file_channel)
{
  return _set_source_member(id, &Source::audio_file_channel, file_channel);
}

bool Scene::set_source_position_fixed(id_t id, const bool& fixed)
{
  return _set_source_member(id, &Source::fixed_position, fixed);
}

bool Scene::set_source_file_length(id_t id, const long int& length)
{
  return _set_source_member(id, &Source::file_length, length);
}

void Scene::set_reference_position(const Position& position)
{
  _reference.position = position;
}

void Scene::set_reference_orientation(const Orientation& orientation)
{
  _reference.orientation = orientation;
}

bool Scene::set_source_model(id_t id, const Source::model_t& model)
{
  return _set_source_member(id, &Source::model, model);
}

// linear volume!
void Scene::set_master_volume(float volume)
{
  _master_volume = volume;
}

/// linear volume!
void Scene::set_master_volume_correction(float corr_factor)
{
  _master_volume_correction = corr_factor;
}

void Scene::set_amplitude_reference_distance(float dist)
{
  _amplitude_reference_distance = dist;
}

// linear scale!
void Scene::set_master_signal_level(float level)
{
  _master_signal_level = level;
}

void Scene::set_cpu_load(float load)
{
  _cpu_load = load;
}

void Scene::set_sample_rate(unsigned int sr)
{
  _sample_rate = sr;
}


void Scene::clear_all_active_sources()
{
  for (loudspeakers_t::iterator i = _loudspeakers.begin();
      i != _loudspeakers.end(); ++i)
  {
    i->active_source_ids.clear();
  }
}

void Scene::set_active_sources(
    Loudspeaker::container_t::size_type loudspeaker_id,
    const std::set<id_t>& active_source_ids)
{
  if (loudspeaker_id <= _loudspeakers.size())
  {
    // loudspeaker IDs start with 1, vector indices with 0!
    _loudspeakers.at(loudspeaker_id - 1).active_source_ids = active_source_ids;
    // we could also swap() the two sets, but that would be premature
    // optimization and the function would need a non-const reference argument.
    // And it would need the precondition that clear_all_active_sources() is
    // called before.
  }

}

void Scene::set_processing_state(ssr::processing_state_t state)
{
  _processing_state = state;
}

/// _. @return processing state
ssr::processing_state_t Scene::get_processing_state() const
{
  return _processing_state;
}

/// _. @return master volume
float Scene::get_master_volume() const
{
  return _master_volume;
}

/** This function returns the master volume with the 
 * correction factor considered. This value is used to
 * adopt the replay volume to a given loudspeaker system.
 * Use it in the audio processing functions only. 
 * @return corrected master volume in linear scale
 */
float Scene::get_corrected_master_volume() const
{
  return _master_volume_correction * this->get_master_volume();
}

/// _. @return amplitude reference distance
float Scene::get_amplitude_reference_distance() const
{
  return _amplitude_reference_distance;
}

/// _. @return instantaneous overall audio signal level
float Scene::get_master_signal_level() const
{
  return _master_signal_level;
}

/// _. @return CPU load in percent
float Scene::get_cpu_load() const
{
  return _cpu_load;
}

/// _. @return current sample rate
unsigned int  Scene::get_sample_rate() const
{
  return _sample_rate;
}


//void Scene::set_transport_state(JackClient::State state)
void Scene::set_transport_state(const std::pair<bool, jack_nframes_t>& state)
{
  //_transport_state = state;
  _transport_playing = state.first;
  _transport_position = state.second;
}

bool Scene::is_playing() const
{
  //return _transport_state.playing;
  return _transport_playing;
}

jack_nframes_t Scene::get_transport_position() const
{
  //return _transport_state.position;
  return _transport_position;
}

Source Scene::get_source(id_t id) const
{ 
  Source source;
  const Source* const source_ptr = maptools::get_item(_source_map, id);

  if (!source_ptr) ERROR("Source " << id << " doesn't exist!");
  else source = *source_ptr;

  return source;
}

/** _.
 * @param id ID of the source
 * @return position of the source
 * @warning If @a id is not found, a auto_ptr to NULL is returned!
 **/
std::auto_ptr<Position> Scene::get_source_position(id_t id) const
{
  std::auto_ptr<Position> position(new Position); // standard ctor
  if (_get_source_member(id, &Source::position, *position)) return position;
  position.reset(); // set to NULL
  return position;
}

/** _.
 * @param id ID of the source
 * @return position of the source
 * @warning If @a id is not found, a auto_ptr to NULL is returned!
 **/
std::auto_ptr<Orientation> Scene::get_source_orientation(id_t id) const
{
  std::auto_ptr<Orientation> orientation(new Orientation); // standard ctor
  if (_get_source_member(id, &Source::orientation, *orientation)) return orientation;
  orientation.reset(); // set to NULL
  return orientation;
}

/** _.
 * @param id ID of the source
 * @return source model
 * @warning If @a id is not found, Source::unknown is returned
 **/
Source::model_t Scene::get_source_model(id_t id) const
{
  Source::model_t model;
  if (_get_source_member(id, &Source::model, model)) return model;
  return Source::unknown;
}

/** _.
 * @param id ID of the source
 * @return linear source gain (=volume)
 * @warning If @a id is not found, 0 is returned
 **/
float Scene::get_source_gain(id_t id) const
{
  float gain;
  if (_get_source_member(id, &Source::gain, gain)) return gain;
  return 0;
}

/** _.
 * @param id ID of the source
 * @return mute state
 * @warning If @a id is not found, false is returned
 **/
bool Scene::get_source_mute_state(id_t id) const
{
  bool state;
  if (_get_source_member(id, &Source::mute, state)) return state;
  return false;
}

/** _.
 * @param id ID of the source
 * @return source name
 * @warning If @a id is not found, an empty string is returned
 **/
std::string Scene::get_source_name(id_t id) const
{
  std::string name;
  if (_get_source_member(id, &Source::name, name)) return name;
  return std::string("");
}

bool Scene::get_source_position_fixed(id_t id) const
{
  bool boolean;
  if (_get_source_member(id, &Source::fixed_position, boolean)) return boolean;

  return false;
}

std::string Scene::get_source_properties_file(id_t id) const
{
  std::string name;
  if (_get_source_member(id, &Source::properties_file, name)) return name;
  return std::string("");
}

Scene::PrivateLoudspeaker::PrivateLoudspeaker(const Loudspeaker& other) :
  Loudspeaker(other)
{
  this->active = false; // the this field is invalid in PrivateLoudspeaker!
  // instead, the std::set active_sources is used.
}

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
