/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * %WFSRenderer class (implementation).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#include <cassert>       // for assert()
#include <cmath>         // for cos()
#include "wfsrenderer.h"
#include "controller.h"  // for set_active_sources(), ...
#include "ssr_global.h"
#include "maptools.h"    // for purge()
#include "mathtools.h"   // for PI
#include "ptrtools.h"    // for is_null(), destroy()
#include "stringtools.h" // for A2S()
#include "audiofileiotools.h"

using ptrtools::is_null;
using ptrtools::destroy;
using stringtools::A2S;
using namespace ssr; // for id_t, processing, ready, ...

/** _.
 * @param scene Reference to the Scene object of the calling class. This
 * is used to obtain the geometry data in each JACK callback.
 * @param controller Reference to the Controller object. This is used to send
 * back information about active/inactive loudspeakers.
 * @param delayline_size Size of the delay lines in samples (all of them have
 * the same length).
 * @param initial_delay Amount of negative delay that can be used when reading
 * from the delay line
 * @param client_name see JackClient::create()
 * @return auto_ptr to the new WFSRenderer object.
 * @todo information about active loudspeakers per source should be passed to
 * the Controller.
 **/
WFSRenderer::ptr_t WFSRenderer::create(const Scene& scene,
    Controller& controller, const std::string& prefilter_file_name,
    const delay_size_t delayline_size,
    const delay_size_t initial_delay, const std::string& client_name)
{
  ptr_t temp; // temp = NULL
  try
  {
    temp.reset(new WFSRenderer(scene, controller, prefilter_file_name,
          delayline_size, initial_delay, client_name));
  }
  catch(jack_error& e)
  {
    ERROR(e.what());
  }
  catch(wfs_error& e)
  {
    ERROR(e.what());
  }
  return temp;
}

/** ctor.
 * -# get number of loudspeakers from the scene and create outputs.
 * -# load WFS pre-filter
 * -# resize the silence buffer
 * -# and, finally, set the Renderer::_client_state to "ready"
 * @param scene see create()
 * @param controller see create()
 * @param delayline_size see create()
 * @param initial_delay see create()
 **/
WFSRenderer::WFSRenderer(
    const Scene& scene,
          Controller&  controller,
    const std::string& prefilter_file_name,
    const delay_size_t delayline_size,
    const delay_size_t initial_delay,
    const std::string& client_name) throw (jack_error, wfs_error) :
  Renderer(scene, client_name, true), // base class ctor.
  delayline_size(delayline_size),
  initial_delay(initial_delay),
  _silence(this->get_buffer_size()), // fill silence buffer with zeroes
  _controller(controller),
  _input_prefix("in"),
  _output_prefix("out"),
  _prefilter_file_name(prefilter_file_name)
{
  if (!_make_outputs(_scene.get_number_of_loudspeakers()))
  {
    maptools::purge(_output_map); // delete any leftovers in _output_map
    throw(wfs_error("Output objects could not be created!"));
  }

  if (prefilter_file_name.empty())
  {
    _prefilter.clear();
    WARNING("No WFS prefilter specified. I won't prefilter.");
  }
  // check prefilter
  else 
  {
    SF_INFO info_container;
    
    if (!audiofileiotools::get_audio_file_info(prefilter_file_name, info_container))
    {
      throw(wfs_error("Can not load WFS prefilter from file '"
		      + prefilter_file_name +"'!"));
    }
    else if (info_container.channels != 1)
    {

      throw(wfs_error("WFS prefilter in file '" + prefilter_file_name 
		      + "' does not have one single channel. It has " 
		      + A2S(info_container.channels) + "."));  
    }
    else if (static_cast<unsigned int>(info_container.samplerate) != this->get_sample_rate())
    {
      throw(wfs_error("WFS prefilter in file '" + prefilter_file_name 
		      + "' has a sample rate of " 
		      + A2S(info_container.samplerate) 
		      + ". JACK sample rate is " 
		      + A2S(this->get_sample_rate())));
    }
    // load impulse response for prefilter
    else if (!audiofileiotools::load_filter_t(prefilter_file_name, _prefilter))
    {
      throw(wfs_error("Can not load WFS prefilter from file '"
		      + prefilter_file_name +"'!"));
    }
    else
    { 
      VERBOSE("Using '" + prefilter_file_name +"' as prefilter for WFS.");
    }
    
  }

  // check if "negative" delay is sufficient for the loudspeaker configuration
  Loudspeaker::container_t test_ls;
  _scene.get_loudspeakers(test_ls, false); // get relative positions
  float max_length = 0; // maximum distance from a loudspeaker to reference
  for (Loudspeaker::container_t::const_iterator i = test_ls.begin();
      i != test_ls.end(); ++i)
  {
    max_length = std::max(i->position.length(), max_length);
  }
  VERBOSE2("Farthest loudspeaker is located " << max_length
      << " meters from reference.");
  // TODO: use the maximum distance as initial delay?
  float compensated_length = ssr::c * initial_delay / this->get_sample_rate();
  VERBOSE2("WFSRenderer is compensating the delay for " << compensated_length
      << " meters (= " << initial_delay << " samples).");
  VERBOSE2("By the way, the buffersize is " << this->get_buffer_size()
      << " samples.");

  // There shouldn't be any sources in the scene at this point.
  // If there are, they are ignored!

  if (this->get_buffer_size() == 0)
  {
    WARNING("Buffersize == 0!");
  }

  _controller.publish_sample_rate((unsigned int)this->get_sample_rate());
  _controller.set_processing_state(ready); // TODO: lock with mutex?

  this->activate(); // activate JACK client
}

WFSRenderer::~WFSRenderer()
{
  _controller.set_processing_state(exiting);
  // cond_signal to any waiting threads?
  maptools::purge(_output_map);
  VERBOSE2("WFSRenderer: Output map cleared.");
  maptools::purge(_input_map);
  VERBOSE2("WFSRenderer: Input map cleared.");
}

/** stop rendering.
 * This is only called by JACK. Do not dare to call it yourself!
 * - stops rendering
 * - calls its base class counterpart
 *
 * @return I don't know what
 */
int WFSRenderer::jack_shutdown()
{
  ERROR("Prematurely disconnected from JACK server! "
      "Trying to stop processing ...");
  _controller.set_processing_state(exiting);
  // cond_signal to any waiting threads?
  maptools::purge(_output_map); WARNING("Output map cleared.");
  maptools::purge(_input_map);  WARNING("Input map cleared.");

  return JackClient::jack_shutdown(); // call original base class function
}

bool WFSRenderer::new_source(id_t id, const std::string& port_name)
{
  if (_input_map.find(id) != _input_map.end())
  {
    WARNING("Input '" << id << "' already exists!");
    return false;
  }
  Input::ptr_t temp(Input::create(*this,
        _scene,
        _controller,
        id,
        _input_prefix + "_" + A2S(id),
        _prefilter));

  if (is_null(temp))
  {
    WARNING("Initialization of input '" << id << "' failed!");
    return false;
  }

  // connect source port with input port
  if (!this->connect_ports(port_name, temp->port_name))
  {
    return false;
  }

  _input_map[id] = temp.release();

  VERBOSE("WFSRenderer: Input " << id << " registered.");
  return true;
}

bool WFSRenderer::new_source(id_t id, const std::string& port_name, 
			     const std::string& brir_file)
{
  (void) id;
  (void) port_name;
  (void) brir_file;
  
  ERROR("This function may not be called.");
  
  return false;
}

/** _.
 * This will only be called once. Loudspeakers can @b not be added later.
 * %Loudspeaker-IDs always range from 1 to @a number.
 * @param number number of loudspeakers
 * @return @b true on success.
 **/
bool WFSRenderer::_make_outputs(Scene::loudspeakers_t::size_type number)
{
  if (!_output_map.empty())
  {
    WARNING("BUG: Output map is not empty!");
    maptools::purge(_output_map);
  }
  for (Scene::loudspeakers_t::size_type id = 1; id <= number; ++id)
  {
    Output::ptr_t temp(Output::create(*this, _output_prefix + "_" + A2S(id)));
    if (is_null(temp))
    {
      WARNING("Initialization of output " << id << " failed!");
      maptools::purge(_output_map);
      return false;
    }
    _output_map[id] = temp.release();
  }
  VERBOSE("WFSRenderer: " << number << " Output objects registered.");
  return true;
}

bool WFSRenderer::delete_source(const id_t id)
{
  return maptools::delete_item(_input_map, id);
}


void WFSRenderer::delete_all_sources()
{
  maptools::purge(_input_map);
  // TODO: anything else?
}

/** _.
 * @param id identifier of the input channel
 * @return Pointer to the corresponding @a id. If none was found, @b NULL is
 * returned.
 **/
const WFSRenderer::Input* WFSRenderer::get_input(id_t id) const
{
  return maptools::get_item(_input_map, id);
}

/** fill a buffer with zeroes.
 * @param buffer pointer to the memory to be cleared
 * @param buffer_size number of data elements to be cleared
 **/
void WFSRenderer::clear_buffer(sample_t* const buffer,
    const jack_nframes_t buffer_size) const
{
  if (buffer_size != _silence.size())
  {
    // this should never happen because JACK buffer size will not change!
    WARNING("BUG?: silence resized! (" << _silence.size() << " -> "
        << buffer_size << ")");
    _silence.resize(buffer_size, 0);
  }
  std::copy(_silence.begin(), _silence.end(), buffer);
}

void WFSRenderer::get_output_port_names(port_list_t& port_list) const
{
  for (output_map_t::const_iterator i = _output_map.begin();
      i != _output_map.end(); ++i)
  {
    assert(i->second != NULL);
    port_list.push_back(i->second->port_name);
  }
}

/** JACK process callback function.
 * This function is called every time the JACK server has new audio data.
 * @param nframes number of frames
 * @return I don't know what
 */
int WFSRenderer::jack_process(const jack_nframes_t nframes)
{
// code must be suitable for real-time execution.
// That means following function are not allowed:
// malloc, free, printf, pthread_mutex_lock, sleep,
// wait, poll, select, pthread_join, pthread_cond_wait,
// etc, etc.

  // clear all information about active/inactive loudspeakers
  _controller.clear_all_active_sources();
  // send some information to the controller (and then to the subscribers)
  _controller.set_transport_state(this->get_transport_state());
  _controller.set_cpu_load(this->get_cpu_load());

  switch (_scene.get_processing_state())
  {
    case processing:
      {
        // write new audio data into all delaylines
        for (input_map_t::const_iterator i = _input_map.begin();
            i != _input_map.end(); ++i)
        {
          assert(i->second != NULL);
          i->second->fetch_new_buffer(nframes);
        }

        Loudspeaker::container_t loudspeakers;
        // necessary memory is reserved inside the function
        _scene.get_loudspeakers(loudspeakers);
        // the loudspeakers contain already absolute positions

        // TODO: remove this if clause?
        if (loudspeakers.size() != _output_map.size())
        {
          WARNING("Different number of loudspeakers in "
              "Scene and WFSRenderer!");
          // TODO: how to react on that?
        }

        // update loudspeaker properties in output opjects
        for (Loudspeaker::container_t::size_type i = 0;
            i < loudspeakers.size(); ++i)
        {
          // WARNING: vector indices start with 0, loudspeaker IDs start with 1!
          if (_output_map.find(i + 1) == _output_map.end())
          {
            // This should never happen!
            ERROR("Loudspeaker #" << i + 1 << " not found in output map!");
          }
          else *_output_map[i + 1] = loudspeakers[i]; // assign new position,...
          // WFSRenderer::Output::operator=() is called!
        }

        // TODO: is it more efficient to initialize the list as member variable
        // and clear it at each jack_process?
        SourceCopy::container_t sources;
        // reservation of memory is done in Scene::get_sources
        // if a std::vector is used (like in our case)
        _scene.get_sources(sources);

        // SET FOCUSSED/NON-FOCUSSED SOURCES

        for (SourceCopy::container_t::iterator source = sources.begin();
            source != sources.end(); ++source)
        {
          // focusedness is irrelevant for plane waves
          if (source->model == Source::plane) continue;

          source->focused = true;
          for (Loudspeaker::container_t::const_iterator loudspeaker
              = loudspeakers.begin(); loudspeaker != loudspeakers.end();
              ++loudspeaker)
          {
            // angle (modulo) between the line connecting source<->loudspeaker
            // and the loudspeaker orientation
            float a = fabs(fmod(angle(loudspeaker->position - source->position,
                loudspeaker->orientation), 2 * mathtools::PI));
            // subwoofers have to be ignored!
            if (loudspeaker->model != Loudspeaker::subwoofer
                && (a < mathtools::PI/2 || a > (mathtools::PI * 3)/2))
            {
              // if at least one loudspeaker "turns its back" to the source, the
              // source is considered non-focused
              source->focused = false;
              break;
            }
          }
        }

        const DirectionalPoint reference = _scene.get_reference();
        float master_volume = _scene.get_corrected_master_volume();

        // splitting the loudspeaker list for several threads will be HERE!

        float master_level = 0.0f;

        // iterate through all output objects via output_map
        for (output_map_t::const_iterator i = _output_map.begin();
            i != _output_map.end(); ++i)
        {
          assert(i->second != NULL);
          i->second->render(sources, reference, nframes, master_volume,
			    _scene.get_amplitude_reference_distance());
          master_level = std::max(master_level,
                i->second->get_signal_level(nframes));

          if (i->second->active_source_ids.size() != 0)
          {
            _controller.set_active_sources(i->first,
                i->second->active_source_ids);
          }
        }

	// in case they move
	_store_source_positions(sources);

	// for GUI
        _controller.set_master_signal_level(master_level);

        // return something?
        break;
      }

    case ready: // audio processing is OFF
      // clear all output buffers
      for (output_map_t::const_iterator i = _output_map.begin();
          i != _output_map.end(); ++i)
      {
        assert(i->second != NULL); // assert that object really exists
        i->second->write_empty_buffer(nframes);
      }
      break;

    case initializing:
      WARNING("JACK process callback called while in \"initializing\" state!");
      // return something bad?
      break;

    case exiting:
      WARNING("JACK process callback called while in \"exiting\" state!");
      // return something bad?
      break;

    default:
      ERROR("BUG: Unknown client state!");
      // return something bad?
  }
  // TODO: find out what the return value means
  return 0;
}

/** This function stores the source positions of the previous
 * audio frame. This information is used in order to properly
 * handle moving sources.
 **/
void 
WFSRenderer::_store_source_positions(SourceCopy::container_t& sources)
{
  _previous_source_positions.clear();

  for (SourceCopy::container_t::iterator source = sources.begin();
       source != sources.end(); ++source)
  {
    _previous_source_positions[source->id] = 
      Position(source->position);
  }
}

unsigned int WFSRenderer::get_number_of_output_channels() const
{
  return _output_map.size();
}

Renderer::renderer_t WFSRenderer::get_renderer_type() const
{
  return Renderer::WFS;
}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class WFSRenderer::SourceCopy
////////////////////////////////////////////////////////////////////////////////

WFSRenderer::SourceCopy::SourceCopy(const std::pair<id_t, Source>& other) :
  DirectionalPoint(other.second), // copy ctor of base class
  id(other.first),
  model(other.second.model),
  gain(other.second.gain),
  mute(other.second.mute)
{}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class WFSRenderer::Output
////////////////////////////////////////////////////////////////////////////////

/** ctor.
 * create a JACK output port and get its full name.
 * @param renderer calling WFSRenderer object
 * @param port_name the desired output port name (the part after the ":")
 *
 * The full port name including the client name can be obtained from the
 * port_name member variable.
 * @throw output_error
 **/
WFSRenderer::Output::Output(const WFSRenderer& renderer,
    const std::string& port_name_) throw (output_error) :
  _renderer(renderer),
  _port(_renderer.register_out_port(port_name_)),
  // get actual port name and save it to member variable
  port_name(_port ? jack_port_name(_port) : "")
{
  if (_port == NULL)
  {
    throw output_error("Output port '" + port_name_
        + "' could not be registered.");
  }
}

WFSRenderer::Output::~Output()
{
  if (_port) _renderer.unregister_port(_port);
}

WFSRenderer::Output::ptr_t WFSRenderer::Output::create(
    const WFSRenderer& renderer, const std::string& port_name)
{
  ptr_t temp; // temp = NULL
  try
  {
    temp.reset(new Output(renderer, port_name));
  }
  catch(output_error& e)
  {
    ERROR(e.what());
  }
  return temp;
}

WFSRenderer::Output& WFSRenderer::Output::operator=(const Loudspeaker& ls)
{
  this->DirectionalPoint::operator=(ls);
  this->model  = ls.model;
  this->weight = ls.weight;
  this->mute   = ls.mute;
  this->active = ls.active; // even if we are not interested in it
  return *this;
}

float WFSRenderer::Output::get_signal_level(const jack_nframes_t nframes)
{
  // get JACK output buffer
  sample_t* out;
  out = static_cast<sample_t*>(jack_port_get_buffer(_port, nframes));

  return _renderer.get_max_signal_amplitude(out, nframes);
}

void WFSRenderer::Output::write_empty_buffer(const jack_nframes_t nframes) const
{
  if (_port)
  {
    sample_t *out;
    out = static_cast<sample_t*>(jack_port_get_buffer(_port, nframes));
    _renderer.clear_buffer(out, nframes);
  }
}

/** Iterates thru the given sources, checks if they
 * are relevant. Calculates delays and weighting factors, reads from the
 * corresponding delay lines and writes to the JACK output buffer.
 * The audio samples are multiplied by the gain factor set in the Input object.
 * @param source_list As the name suggests, a list of sources
 * @param nframes Number of samples to be calculated in one block
 *
 * \par How to calculate delay and weighting factor
 * The delay and the weighting factor are calculated as described in the
 * diploma thesis of Dirk Opperschall ("Realisierung eines Demonstrators fuer
 * Punktquellen und ebene Wellen fuer ein Wellenfeldsynthese-System",
 * Universit&auml;t Erlangen, 2002).
 * \par Calculating the delay
 * \f$ \displaystyle \tau_{n,k} = -\lambda \frac{r_{n,k}}{c} \f$
 * \par Calculating the weighting factor
 * \f$ \displaystyle a_{n,k} = \frac{g_0(\mathbf{r}_{n,k}) \cos
 * \phi_{inc,n,k}}{\sqrt{r_{n,k}}} \f$
 * with
 * \f$ \displaystyle
 * g_0(\mathbf{r}_{L,n}) = \sqrt{\frac{\Delta
 * r_0(\mathbf{r}_{L,n})}{r_0(\mathbf{r}_{L,n}) - \lambda\Delta
 * r_0(\mathbf{r}_{L,n})}}
 * \f$
 * and \f$ \Delta r_0(\mathbf{r}_{L,n}) = |\mathbf{r}_{R,0}(\mathbf{r}_{L,n})
 * - \mathbf{r}_{L,n}| \f$.
 * \par Explanation of the mathematical Symbols
 * The loudspeaker array consists of \f$N\f$ loudspeakers numbered by \f$n\f$
 * from 1 to \f$N\f$. There are \f$K\f$ sources numbered by \f$k\f$ from 1 to
 * \f$K\f$.
 * \par
 * \f$\lambda = 1\f$ for focussed sources, \f$\lambda = -1\f$ for non-focussed
 * sources.
 * 
 * [to be completed ...]
 * 
 * \par plane waves
 * In the case of plane waves the sound pressure does not decrease with
 * increasing distance. So the weighting factor \f$a_n\f$ for the
 * \f$n^\mathrm{th}\f$
 * loudspeaker only depends on the angle \f$\phi_{inc,n}\f$ between the
 * propagation direction of the plane wave and the "direction" of the
 * loudspeaker (the normal vector of the secondary source distribution at the
 * position of the loudspeaker).
 * \par
 * The weighting factor is calculated by \f$ a_n = \cos \phi_{inc,n} \f$.
 * \par
 * The delay is calculated similarly to the point source,
 * \f$\displaystyle \tau_n = \frac{r_n}{c} \f$, where \f$r_n\f$ is the (closest)
 * distance between the plane wave and the loudspeaker and \f$c\f$ is the speed
 * of sound.
 **/
void WFSRenderer::Output::render(const SourceCopy::container_t& sources,
    const DirectionalPoint& reference, const jack_nframes_t nframes,
    const float master_volume, const float ampl_ref_dist)
{
  // get JACK output buffer
  sample_t* out;
  out = static_cast<sample_t*>(jack_port_get_buffer(_port, nframes));

  VERBOSE3("\n======================================"
      "==========================================");
  VERBOSE3("== output position: x: " << this->position.x << ", y: "
      << this->position.y << ", azimuth: " << this->orientation.azimuth);
  VERBOSE3("== JACK port: " << this->port_name);

  // the first output block replaces the contents of "out", then "accumulate" is
  // set to "true".
  bool accumulate = false;

  this->active_source_ids.clear(); // reset active sources to an empty set

  std::map<id_t, Coefficient> current_coefficients; // empty map

  float reference_distance = (this->position - reference.position).length();

  for (SourceCopy::container_t::const_iterator i = sources.begin();
      i != sources.end(); ++i)
  {
    if (i->mute == true) continue; // next source

    float weighting_factor = 0.0f;
    float delay            = 0.0f;
    // input object must not be deleted! It's deleted in ~WFSRenderer()
    const Input* current_input = _renderer.get_input(i->id);
    if (current_input == NULL)
    { // this happens regularly when a source is added and the renderer is
      // already processing because it takes some time until the JACK ports are
      // ready. Therefore the following warning is only printed if verbose > 1.
      VERBOSE2("WFSRenderer::Output: Input for source " << i->id
          << " doesn't exist in input map!");
      continue; // continue with next source
    }

    // define a restricted area around loudspeakers to avoid division by zero:
    const float safety_radius = 0.01f; // 1 cm

    switch (i->model) // check if point source or plane wave or ...
    {
      case Source::point:
        if (this->model == Loudspeaker::subwoofer)
        {
          // the delay is calculated to be correct on the reference position
          // delay can be negative!
          delay = (i->position-reference.position).length()-reference_distance;

          if (fabs(delay) < safety_radius)
          {
            weighting_factor = 1.0f / sqrt(safety_radius);
          }
          else
          {
            weighting_factor = 1.0f / sqrt(fabs(delay));
          }
          VERBOSE3("point source (for subwoofer): id: " << i->id);
          break; // step out of switch
        }

        delay = (this->position - i->position).length();
        assert(delay >= 0);

        float denominator;
        if (delay < safety_radius) denominator = sqrt(safety_radius);
        else denominator = sqrt(delay);

        // TODO: does this really do the right thing?
        weighting_factor = cos(angle(this->position - i->position,
              this->orientation)) / denominator;

        // TODO: factor with reference point/reference line is missing!

        if (weighting_factor < 0.0f)
        {
          // negative weighting factor is only valid for focused sources
          if (i->focused)
          {
            // loudspeaker selection:

            // this could also be done by using the cosine function instead of
            // inner product

            // calculate inner product of those two vectors:
            // this = loudspeaker; i = source
            Position ls =     this->position - i->position;
            Position rs = reference.position - i->position;

            // TODO: write inner product function in Position class
            if ((ls.x * rs.x + ls.y * rs.y) < 0.0f)
            {
              // if the inner product is less than zero, the source is more or
              // less between the loudspeaker and the reference
              delay = -delay;
              weighting_factor = -weighting_factor;
              VERBOSE3("point source (focused): id: " << i->id
                  << ", factor: " << weighting_factor);
            }
            else
            {
              VERBOSE3("point source (focused but ignored): id: " << i->id
                  << ", factor: " << weighting_factor);
              continue; // continue with next source
            }
          }
          else // non-focused and weighting_factor < 0
          {
            VERBOSE3("point source (non-focused but ignored): id: " << i->id
                << ", factor: " << weighting_factor);
            continue; // continue with next source
          }
        }
        else if(weighting_factor > 0.0f) // positive weighting factor
        {
          // negative weighting factor is only valid for focused sources
          if (!i->focused)
          {
            VERBOSE3("point source (non-focused): id: " << i->id
                << ", factor: " << weighting_factor);
          }
          else // focused
          {
            VERBOSE3("point source (focused but ignored): id: " << i->id
                << ", factor: " << weighting_factor);
            continue; // continue with next source
          }
        }
        else
        {
          // this should never happen:
          VERBOSE3("Weighting factor is 0 or +-Inf or NaN!");
          continue; // continue with next source
        }
        break;

      case Source::plane:
        if (this->model == Loudspeaker::subwoofer)
        {
          weighting_factor = 1.0f; // TODO: is this correct?
          // the delay is calculated to be correct on the reference position
          // delay can be negative!
          delay = i->plane_to_point_distance(reference.position)
            - reference_distance;
          VERBOSE3("plane wave (for subwoofer): id: " << i->id);
          break; // step out of switch
        }

        // weighting factor is determined by the cosine of the angle
        // difference between plane wave direction and loudspeaker direction
        weighting_factor = cos(angle(*i, *this));
        // check if loudspeaker is active for this source
        if (weighting_factor < 0)
        {
          VERBOSE3("plane wave   (ignored): id: " << i->id << ", factor: "
              << weighting_factor);
          continue; // continue with next source
        }

        delay = i->plane_to_point_distance(this->position);

        if (delay < 0.0)
        {
          VERBOSE3("plane wave (\"focussed\"): id: " << i->id
              << ", factor: " << weighting_factor);
        }
        else // positive delay
        {
          VERBOSE3("plane wave: id: " << i->id << ", factor: "
              << weighting_factor);
        }
        break;

      default:
        WARNING("Unknown source model");
        continue; // continue with next source
    } // switch source model

    // no distance attenuation for plane waves 
    if (i->model == Source::plane)
    {
      weighting_factor *= 0.5f / ampl_ref_dist; // 1/r
      // weighting_factor *= 0.25f / sqrt(ampl_ref_dist); // 1/sqrt(r)
    }
    else
    {
      // consider distance attenuation
      float source_distance = (i->position - reference.position).length();

      // no volume increase for sources closer than 0.5m to reference position
      source_distance = std::max(source_distance, 0.5f);

      weighting_factor *= 0.5f / source_distance; // 1/r
      // weighting_factor *= 0.25f / sqrt(source_distance); // 1/sqrt(r)
    }

    // apply the gain factor of the current source
    weighting_factor *= i->gain;

    // apply tapering
    weighting_factor *= this->weight;

    // apply master volume
    weighting_factor *= master_volume;

    assert(weighting_factor >= 0.0f);

    VERBOSE3_NOLF("   delay = " << delay << " meters = ");
    // delay in seconds
    delay *= c_inverse;
    VERBOSE3_NOLF(delay << " seconds = ");

    // delay in samples
    delay *= _renderer.get_sample_rate();

    VERBOSE3(delay << " samples.");

    // TODO: check for negative delay and print an error if > initial_delay

    // TODO: do proper rounding
    // TODO: enable interpolated reading from delay line.
    int int_delay = static_cast<int>(delay + 0.5f);


    int previous_int_delay = int_delay;
    float previous_weighting_factor = 0; // fade in

    std::map<id_t, Coefficient>::iterator iter = _previous_coefficients.find(i->id);

    if (iter != _previous_coefficients.end())
    {
      previous_int_delay = iter->second.delay;
      previous_weighting_factor = iter->second.factor;
    }

    // read from corresponding delay line, multiply and add to output buffer.
    // IMPORTANT: first source is not added but replaces the buffer content.

    /*
    if (!current_input->read_mult(out, nframes, int_delay, weighting_factor,
      accumulate))
    */

    if (!current_input->read_fade(out, nframes, previous_int_delay,
          previous_weighting_factor, int_delay, weighting_factor, accumulate))

    //    if (!current_input->read_interp(out, nframes, previous_int_delay,
    //				    previous_weighting_factor, int_delay, 
    //				    weighting_factor, accumulate))
    {
      ERROR("render(): Output buffer could not be written for source "
          << i->id << "!");
      continue; // continue with next source
    }

    current_coefficients[i->id] = Coefficient(int_delay, weighting_factor);

    // after the first (successful) write operation everything is added to the
    // output.
    accumulate = true;
    // The current loudspeaker is active for the current source:
    this->active_source_ids.insert(i->id);
  } // for all sources

  // if there was no successful write operation (maybe even the source list was
  // empty), the old audio data is still in the buffer.
  // In this case, we have to clear it:
  if (!accumulate)
  {
    // fill buffer with zeros
    _renderer.clear_buffer(out, nframes);
    VERBOSE3("No source active. Output buffer cleared.");
  }

  _previous_coefficients.clear();
  _previous_coefficients = current_coefficients; // if nothing was written, 
                                                 // _previous is emptied.
  // TODO: is _previous.swap(current) faster?

  // TODO: return true or something?
}

////////////////////////////////////////////////////////////////////////////////
// Implementation of the nested class WFSRenderer::Input
////////////////////////////////////////////////////////////////////////////////

/**
 * @param renderer reference to the calling WFSRenderer object. Using this, we
 * can access the public members of the WFS renderer.
 **/
WFSRenderer::Input::Input(const WFSRenderer& renderer, const Scene& scene,
        Controller& controller, id_t source_id, const std::string& port_name_,
        const Convolver::data_t& prefilter) throw (input_error):
  DelayLine<sample_t>(renderer.delayline_size, renderer.initial_delay),
  _renderer(renderer),
  _scene(scene),
  _controller(controller),
  _source_id(source_id),
  _port(_renderer.register_in_port(port_name_)),
  _prefilter_cycles(1u),
  // get actual port name and save it to member variable
  port_name(_port ? jack_port_name(_port) : "")
{
  if (_port == NULL)
  {
    throw input_error("Input port '" + this->port_name
        + "' could not be registered.");
  }

  // _convolver is already pointing to NULL

  if (prefilter.size() != 0) // if prefiltering
  {

    // determine the next highest power of 2 above prefilter size
    const unsigned int power_of_2 =
             mathtools::next_power_of_2(prefilter.size());

    const unsigned int convolver_buffer_size = 
                std::min(power_of_2, _renderer.get_buffer_size());
    
    _prefilter_cycles = _renderer.get_buffer_size()/convolver_buffer_size;
    
    if (_prefilter_cycles != 1u)
    {
      VERBOSE("I create " << _prefilter_cycles << " subcycles of size " 
	      << convolver_buffer_size << "for WFS prefiltering.");
    }

    _convolver = Convolver::create(convolver_buffer_size, Convolver::none);
 
    if (is_null(_convolver))
    {
      throw input_error("Initialization of convolver for input port \""
          + this->port_name + "\" failed!");
    }

    // set prefilter
    _convolver->set_filter_t(prefilter);

  } // if prefiltering
}

WFSRenderer::Input::~Input()
{
  if (_port) _renderer.unregister_port(_port);
}

WFSRenderer::Input::ptr_t WFSRenderer::Input::create(
        const WFSRenderer& renderer,
        const Scene& scene,
        Controller& controller,
        const id_t source_id,
        const std::string& port_name,
        const Convolver::data_t& prefilter)
{
  ptr_t temp; // temp = NULL
  try
  {
    temp.reset(new Input(renderer, scene, controller, source_id, port_name,
          prefilter));
  }
  catch(input_error& e)
  {
    ERROR(e.what());
  }
  return temp;
}

bool WFSRenderer::Input::fetch_new_buffer(jack_nframes_t nframes)
{
  sample_t *in;
  in = static_cast<sample_t*>(jack_port_get_buffer(_port, nframes));
  if (in == NULL)
  {
    WARNING("Got no input buffer from JACK!");
    return false;
  }

  if (is_null(_convolver))
  {
    this->prepare_write();

    // don't prefilter, just write to delay line
    if (!this->write(in, nframes))
    {
      WARNING("JACK input buffer could not be written to delay line.");
      return false;
    }

  }
  else // _convolver != NULL
  {
    // prefilter and write to delay line
    if (!_prefilter_signal(in, nframes/_prefilter_cycles))
    {
      WARNING("JACK input buffer could not be written to delay line.");
      return false;
    }
  }

  if (_scene.get_source_mute_state(_source_id))
  {
    // update source signal level
    _controller.set_source_signal_level(_source_id, 0.0f);
  }
  else
  {
    // update source signal level (max amplitude)
    _controller.set_source_signal_level(_source_id,
        _renderer.get_max_signal_amplitude(in, nframes)
        * _scene.get_source_gain(_source_id));
  }
  return true;
}

bool WFSRenderer::Input::_prefilter_signal(WFSRenderer::sample_t *in,
                                           jack_nframes_t nsubframes)
{
  this->prepare_write();

  for (unsigned int cycle = 0; cycle < _prefilter_cycles; cycle++)
  {
    // prefilter and write to delay line
    if (!this->write(_convolver->convolve_signal(in, 1.0), nsubframes))
    {
      return false;
    }

    // move forward in input buffer
    in += nsubframes;
  }

  return true;
}

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
