/******************************************************************************
 * Copyright (c) 2006-2012 Quality & Usability Lab                            *
 *                         Deutsche Telekom Laboratories, TU Berlin           *
 *                         Ernst-Reuter-Platz 7, 10587 Berlin, Germany        *
 *                                                                            *
 * This file is part of the SoundScape Renderer (SSR).                        *
 *                                                                            *
 * The SSR is free software:  you can redistribute it and/or modify it  under *
 * the terms of the  GNU  General  Public  License  as published by the  Free *
 * Software Foundation, either version 3 of the License,  or (at your option) *
 * any later version.                                                         *
 *                                                                            *
 * The SSR is distributed in the hope that it will be useful, but WITHOUT ANY *
 * WARRANTY;  without even the implied warranty of MERCHANTABILITY or FITNESS *
 * FOR A PARTICULAR PURPOSE.                                                  *
 * See the GNU General Public License for more details.                       *
 *                                                                            *
 * You should  have received a copy  of the GNU General Public License  along *
 * with this program.  If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * The SSR is a tool  for  real-time  spatial audio reproduction  providing a *
 * variety of rendering algorithms.                                           *
 *                                                                            *
 * http://tu-berlin.de/?id=ssr                  SoundScapeRenderer@telekom.de *
 ******************************************************************************/

/** @file
 * %WFSRenderer class (definition).
 *
 * $LastChangedDate: 2012-02-01 11:43:25 +0100 (Wed, 01 Feb 2012) $
 * $LastChangedRevision: 1659 $
 * $LastChangedBy: geier.matthias $
 **/

#ifndef WFSRENDERER_H
#define WFSRENDERER_H

#include <map>
#include <set>
#include <vector>

#include "renderer.h"   // base class of WFSRenderer
#include "delayline.h"  // base class of WFSRenderer::Input
#include "convolver.h"
#include "directionalpoint.h" // Position and Orientation

class Controller; // forward declaration

using ssr::id_t;

/** WFS renderer class.
 * This class does the WFS rendering. Audio processing is done with the
 * inherited JACK functionality.
 **/
class WFSRenderer : public Renderer
{
  private:
    class Output; // nested class, defined later
    class Input;  // nested class, defined later

    /** Class to temporarily store selected source information in.
     * By use of a class like this, the desired information for every source can
     * be obtained using the function Scene::get_source_list().
     **/
    struct SourceCopy : DirectionalPoint
    {
      typedef std::vector<SourceCopy> container_t; ///< list of SourceCopys
      /// type conversion constructor
      SourceCopy(const std::pair<id_t, Source>& other);

      id_t id;               ///< unique ID, see Scene::source_map_t
      Source::model_t model; ///< model, see Source
      float gain;            ///< gain, see Source
      bool mute;             ///< mute/unmute, see Source

      /// focused/non-focussed source, gets calculated at each audio block
      bool focused;
    };

  public:
    /// Type for specifying length and indizes of DelayLines
    typedef DelayLine<sample_t>::size_type delay_size_t;

    typedef std::auto_ptr<WFSRenderer> ptr_t;

    /// exception to be thrown by ctor.
    struct wfs_error : public std::runtime_error
    {
      wfs_error(const std::string& s): std::runtime_error(s) {}
    };

    WFSRenderer(const Scene& scene, Controller& controller,
        const std::string& prefilter_file_name,
        delay_size_t delayline_size, delay_size_t initial_delay,
        const std::string& client_name)
      throw (jack_error, wfs_error);

    /// a map which holds all output objects.
    typedef std::map<Loudspeaker::container_t::size_type, Output*> output_map_t;
    /// a map which holds all input objects.
    typedef std::map<id_t, Input*> input_map_t;

    /// "named constructor" for %WFSRenderer objects
    static ptr_t create(const Scene& scene, Controller& controller,
        const std::string& prefilter_file_name,
        delay_size_t delayline_size, delay_size_t initial_delay = 0,
        const std::string& client_name = "WFS-Renderer");

    /// delete all Input/Output objects and clear input and output map.
    virtual ~WFSRenderer();

    // from JackClient:
    virtual int jack_shutdown();
    // from Renderer:
    virtual bool new_source(id_t id, const std::string& port_name);
    virtual bool new_source(id_t id, const std::string& port_name,
			    const std::string& brir_file);
    virtual bool delete_source(id_t id);
    virtual void delete_all_sources();
    virtual void get_output_port_names(port_list_t& port_list) const;
    virtual unsigned int get_number_of_output_channels() const;
    virtual renderer_t get_renderer_type() const;

    /// Get pointer to input object corresponding to @a id.
    const Input* get_input(id_t id) const;

    void clear_buffer(sample_t* const buffer, jack_nframes_t buffer_size) const;

    const delay_size_t delayline_size; ///< Size of delay lines (in samples)
    const delay_size_t initial_delay;  ///< Initial delay (in samples, >= 0)

  private:
    /// overwrites virtual member function declared in JackClient class
    virtual int jack_process(jack_nframes_t nframes);

    /// this will be filled with zeroes
    mutable std::vector<sample_t> _silence;

    input_map_t   _input_map;    ///< all input  objects are stored here
    output_map_t _output_map;    ///< all output objects are stored here

    Controller& _controller;          ///< Reference to the Controller
    const std::string  _input_prefix; ///< prefix for the input  ports
    const std::string _output_prefix; ///< prefix for the output ports

    const std::string _prefilter_file_name; /// < dto.

    std::map<id_t, Position> _previous_source_positions; ///< source positions during the previous frame

    /// initialize Output objects
    bool _make_outputs(Loudspeaker::container_t loudspeakers);
    bool _make_outputs(Scene::loudspeakers_t::size_type number);

    void _store_source_positions(SourceCopy::container_t& sources);

    Convolver::data_t _prefilter; ///< container for impulse response

};

////////////////////////////////////////////////////////////////////////////////
// Definition of the nested class Output
////////////////////////////////////////////////////////////////////////////////

/** Manages an output buffer and the corresponding JACK calls.
 * @warning z coordinates and elevation are ignored
 **/
class WFSRenderer::Output : public Loudspeaker
{
  public:
    typedef std::auto_ptr<Output> ptr_t;

    /// exception to be thrown by ctor.
    struct output_error : public std::runtime_error
    {
      output_error(const std::string& s): std::runtime_error(s) {}
    };


    /// "named constructor"
    static ptr_t create(const WFSRenderer& renderer,
        const std::string& port_name);

    ~Output(); ///< Unregister associated JACK output port

    /// Render audio data to the corresponding output object.
    void render(const SourceCopy::container_t& source_list,
        const DirectionalPoint& reference, 
        const jack_nframes_t nframes,
	const float master_volume = 1.0f,
	const float ampl_ref_dist = 3.0f);

    float get_signal_level(const jack_nframes_t nframes);

    /// clear output buffer
    void write_empty_buffer(jack_nframes_t nframes) const;

    /// use the assignment operator of the base class.
    using DirectionalPoint::operator=;
    /// overloaded assignment operator for Loudspeaker
    Output& operator=(const Loudspeaker& ls);

    /// sources for which current Output is active
    std::set<id_t> active_source_ids;
    // those are public, because I was too lazy to write get_active_source_ids()

  private:
    const WFSRenderer& _renderer; /// ref. to the calling WFSRenderer object
    Output(const WFSRenderer& renderer, const std::string& port_name)
      throw (output_error);

    jack_port_t* _port;     ///< JACK port associated with this output

    /// to store one delay and one weighting factor
    struct Coefficient
    {
      explicit Coefficient(int int_delay = 0, float weighting_factor = 1.0) :
        delay(int_delay),
        factor(weighting_factor)
      {}
      int   delay;
      float factor;
    };
    std::map<id_t, Coefficient> _previous_coefficients;

    Output(const Output&);            ///< deactivated
    Output& operator=(const Output&); ///< deactivated

  public:
    // another public block for correct initialization order
    const std::string port_name; ///< full name of the JACK port
};

////////////////////////////////////////////////////////////////////////////////
// Definition of the nested class Input
////////////////////////////////////////////////////////////////////////////////

/** Manages an input buffer (including delay line) and the corresponding JACK
 * calls.
 **/
class WFSRenderer::Input : public DelayLine<sample_t>
{
  public:
    typedef std::auto_ptr<Input> ptr_t;

    /// exception to be thrown by ctor.
    struct input_error : public std::runtime_error
    {
      input_error(const std::string& s): std::runtime_error(s) {}
    };

    Input(const WFSRenderer& renderer, const Scene& scene,
          Controller& controller, const id_t source_id,
          const std::string& port_name, const Convolver::data_t& prefilter)
      throw (input_error);

    /// "named constructor"
    static ptr_t create(const WFSRenderer& renderer,
            const Scene& scene,
            Controller& controller,
            const id_t source_id,
            const std::string& name,
            const Convolver::data_t& prefilter);

    ~Input(); ///< Unregister associated JACK input port

    /// get a buffer full of audio from JACK and write it into the delay line
    bool fetch_new_buffer(jack_nframes_t nframes);

  private:
    /// holds a reference to the calling WFSRenderer object
    const WFSRenderer& _renderer;

    const Scene& _scene; ///< holds a reference to the scene

    Controller& _controller; ///< Reference to the Controller

    Convolver::ptr_t _convolver;

    id_t _source_id; ///< id of associated sound source

    jack_port_t* _port; ///< the JACK port corresponding to this input object

    unsigned int _prefilter_cycles; ///< number of subframes for prefiltering

    Input(const Input&);            ///< deactivated
    Input& operator=(const Input&); ///< deactivated

    bool _prefilter_signal(WFSRenderer::sample_t *in,
			   jack_nframes_t nsubframes);

  public:
    const std::string port_name; ///< full name of the JACK port
};

#endif

// Settings for Vim (http://www.vim.org/), please do not remove:
// vim:softtabstop=2:shiftwidth=2:expandtab:textwidth=80:cindent
